from typing import List

import numpy as np

from sonusai.mixture.mixdb import ClassCount
from sonusai.mixture.mixdb import Mixture
from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureID
from sonusai.mixture.mixdb import Mixtures


def get_class_count(mixdb: MixtureDatabase,
                    mixture: Mixture,
                    target_audios: List[np.ndarray],
                    noise_audio: np.ndarray) -> ClassCount:
    """Computes the number of samples for which each truth index is active for a given sample-based truth input."""
    from sonusai.mixture import get_class_weights_threshold
    from sonusai.mixture import generate_truth

    truth = generate_truth(mixdb=mixdb,
                           mixture=mixture,
                           target_audios=target_audios,
                           noise_audio=noise_audio)

    class_weights_threshold = get_class_weights_threshold(mixdb)

    class_count = [0] * mixdb.num_classes
    num_classes = mixdb.num_classes
    if mixdb.truth_mutex:
        num_classes -= 1
    for cl in range(num_classes):
        class_count[cl] = int(np.sum(truth[:, cl] >= class_weights_threshold[cl]))

    return class_count


def get_total_class_count(mixdb: MixtureDatabase, mixid: MixtureID = ':') -> ClassCount:
    """Sums the class counts for all mixtures."""
    from sonusai.mixture import get_mixtures_from_mixid

    mixtures = get_mixtures_from_mixid(mixdb, mixid)
    return compute_total_class_count(mixdb, mixtures)


def compute_total_class_count(mixdb: MixtureDatabase, mixtures: Mixtures) -> ClassCount:
    from sonusai import SonusAIError

    total_class_count = [0] * mixdb.num_classes
    for mixture in mixtures:
        for cl in range(mixdb.num_classes):
            total_class_count[cl] += mixture.class_count[cl]

    if mixdb.truth_mutex:
        # Compute the class count for the 'other' class
        if total_class_count[-1] != 0:
            raise SonusAIError('Error: truth_mutex was set, but the class count for the last count was non-zero.')
        total_class_count[-1] = sum([sub.samples for sub in mixtures]) - sum(total_class_count)

    return total_class_count



from dataclasses import dataclass


@dataclass
class Segment:
    start: int
    length: int

    def get_slice(self) -> slice:
        return slice(self.start, self.start + self.length)

    def trim_start(self, amount: int) -> None:
        self.trim_length(amount)
        self.start += amount

    def trim_length(self, amount: int) -> None:
        if amount >= self.length:
            raise ValueError(f'trim amount greater than or equal to length')
        self.length -= amount


def get_slices(segment: Segment, start: int) -> (slice, slice):
    return segment.get_slice(), slice(start, start + segment.length)



def human_readable_size(size: float, decimal_places: int = 3) -> str:
    """Convert number into string with units"""
    for unit in ['B', 'kB', 'MB', 'GB', 'TB']:
        if size < 1024.0:
            break
        size /= 1024.0
    return f'{size:.{decimal_places}f} {unit}'



from typing import Union

import numpy as np
import pandas as pd

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureID


def snr_summary(mixdb: MixtureDatabase,
                mixid: MixtureID,
                truth_f: np.ndarray,
                predict: np.ndarray,
                segsnr: Union[np.ndarray, None] = None,
                predict_thr: Union[float, np.ndarray] = 0,
                truth_thr: float = 0.5,
                timesteps: int = 0) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame, dict):
    """Calculate average-over-class metrics per SNR over specified mixture list.
       Inputs:
         mixdb        Mixture database
         mixid
         truth_f      Truth/labels (features, num_classes)
         predict      Prediction data / neural net model one-hot out (features, num_classes)
         segsnr       Segmental SNR from SonusAI genft  (transform_frames, 1)
         predict_thr  Decision threshold(s) applied to predict data, allowing predict to be
                      continuous probabilities or decisions
         truth_thr    Decision threshold(s) applied to truth data, allowing truth to be
                      continuous probabilities or decisions
         timesteps

       Default predict_thr=0 will infer 0.5 for multi-label mode (truth_mutex = False), or
       if single-label mode (truth_mutex == True) then ignore and use argmax mode, and
       the confusion matrix is calculated for all classes.

       Returns pandas dataframe (snrdf) of metrics per SNR.
    """
    import warnings

    from sonusai.metrics import one_hot
    from sonusai.mixture import get_mixid_data
    from sonusai.queries import get_mixids_from_snr

    num_classes = truth_f.shape[1]

    snr_mixids = get_mixids_from_snr(mixdb=mixdb, mixid=mixid)

    # Check predict_thr array or scalar and return final scalar predict_thr value
    if not mixdb.truth_mutex and num_classes > 1:
        if np.ndim(predict_thr) == 0 and predict_thr == 0:
            # multi-label predict_thr scalar 0 force to 0.5 default
            predict_thr = 0.5

        if np.ndim(predict_thr) == 1:
            if len(predict_thr) == 1:
                if predict_thr[0] == 0:
                    # multi-label predict_thr array scalar 0 force to 0.5 default
                    predict_thr = 0.5
                else:
                    # multi-label predict_thr array set to scalar = array[0]
                    predict_thr = predict_thr[0]

    macro_avg = np.zeros((len(snr_mixids), 7), dtype=float)
    micro_avg = np.zeros((len(snr_mixids), 7), dtype=float)
    wghtd_avg = np.zeros((len(snr_mixids), 7), dtype=float)
    ssnr_stats = None
    segsnr_f = None

    if segsnr is not None:
        # prep segsnr if provided, transform frames to feature frames via mean()
        # expected to always be an integer
        feature_frames = int(segsnr.shape[0] / truth_f.shape[0])
        segsnr_f = np.mean(np.reshape(segsnr, (truth_f.shape[0], feature_frames)), axis=1, keepdims=True)
        ssnr_stats = np.zeros((len(snr_mixids), 3), dtype=float)

    ii = 0
    for snr in snr_mixids:
        y_truth, y_predict = get_mixid_data(mixdb, snr_mixids[snr], truth_f, predict)
        _, metrics, _, _, _, mavg = one_hot(y_truth, y_predict, predict_thr, truth_thr, timesteps)

        # mavg macro, micro, weighted: [PPV, TPR, F1, FPR, ACC, mAP, mAUC, TPSUM]
        macro_avg[ii, :] = mavg[0, 0:7]
        micro_avg[ii, :] = mavg[1, 0:7]
        wghtd_avg[ii, :] = mavg[2, 0:7]
        if segsnr is not None:
            y_truth, y_segsnr = get_mixid_data(mixdb, snr_mixids[snr], truth_f, segsnr_f)
            with warnings.catch_warnings():
                warnings.filterwarnings(action='ignore', message='divide by zero encountered in log10')
                # segmental SNR mean = mixture_snr and target_snr
                ssnr_stats[ii, 0] = 10 * np.log10(np.mean(y_segsnr))
                # segmental SNR 80% percentile
                ssnr_stats[ii, 1] = 10 * np.log10(np.percentile(y_segsnr, 80, interpolation='midpoint'))
                # segmental SNR max
                ssnr_stats[ii, 2] = 10 * np.log10(max(y_segsnr))

        ii += 1

    # SNR format: PPV, TPR, F1, FPR, ACC, AP, AUC
    col_n = ['PPV', 'TPR', 'F1', 'FPR', 'ACC', 'AP', 'AUC']
    snr_macrodf = pd.DataFrame(macro_avg, index=list(snr_mixids.keys()), columns=col_n)
    snr_macrodf.sort_index(ascending=False, inplace=True)

    snr_microdf = pd.DataFrame(micro_avg, index=list(snr_mixids.keys()), columns=col_n)
    snr_microdf.sort_index(ascending=False, inplace=True)

    snr_wghtdf = pd.DataFrame(wghtd_avg, index=list(snr_mixids.keys()), columns=col_n)
    snr_wghtdf.sort_index(ascending=False, inplace=True)

    # Add segmental SNR columns if provided
    if segsnr is not None:
        ssnrdf = pd.DataFrame(ssnr_stats, index=list(snr_mixids.keys()), columns=['SSNRavg', 'SSNR80p', 'SSNRmax'])
        ssnrdf.sort_index(ascending=False, inplace=True)
        snr_macrodf = pd.concat([snr_macrodf, ssnrdf], axis=1)
        snr_microdf = pd.concat([snr_microdf, ssnrdf], axis=1)
        snr_wghtdf = pd.concat([snr_wghtdf, ssnrdf], axis=1)

    return snr_macrodf, snr_microdf, snr_wghtdf, snr_mixids



import multiprocessing as mp

import numpy as np


def to_numpy_array(arr: mp.Array, dtype=None) -> np.ndarray:
    """Convert multiprocessing array to numpy array"""
    return np.frombuffer(arr.get_obj(), dtype=dtype)



"""sonusai lsdb

usage: lsdb [-ht] [-i MIXID] [-c TID] MIXDB

Options:
    -h, --help
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate and analyze. [default: *].
    -c TID, --truth_index TID       Analyze mixtures that contain this truth index.
    -t, --targets                   List all target files.

List mixture data information from a SonusAI mixture database.

Inputs:
    MIXDB       A SonusAI mixture database JSON file or an HDF5 file containing a SonusAI mixture database.
                The HDF5 file contains:
                    attribute:  mixdb
                    dataset:    feature (optional)
                    dataset:    truth_f (optional)

"""
from typing import Union

import numpy as np

from sonusai import logger
from sonusai.mixture import MixtureDatabase


def lsdb(mixdb: MixtureDatabase,
         mixid: str = '*',
         datafile: str = None,
         truth_index: Union[int, None] = None,
         list_targets: bool = False) -> None:
    import h5py

    from sonusai import SonusAIError
    from sonusai.mixture import calculate_snr_f_statistics
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import get_feature_stats
    from sonusai.mixture import get_truth_indices_for_target
    from sonusai.queries import get_mixids_from_truth_index
    from sonusai.utils import consolidate_range
    from sonusai.utils import expand_range
    from sonusai.utils import print_class_count
    from sonusai.utils import print_mixture_details
    from sonusai.utils import seconds_to_hms

    desc_len = 24

    feature_shape = None
    truth_f_shape = None
    if datafile is not None:
        with h5py.File(datafile, 'r') as f:
            if 'feature' in f.keys():
                feature_shape = f['feature'].shape
            if 'truth_f' in f.keys():
                truth_f_shape = f['truth_f'].shape

    if feature_shape is not None:
        logger.info(f'Feature shape {feature_shape}')

    if truth_f_shape is not None:
        logger.info(f'Truth shape   {truth_f_shape}')

    if feature_shape is not None and truth_f_shape is not None:
        f_frames = feature_shape[0]
        t_frames = truth_f_shape[0]
        if f_frames != t_frames:
            logger.warn(f'Number of feature frames, {f_frames} does not match number of truth frames, {t_frames}.')

    total_samples = sum([sub.samples for sub in mixdb.mixtures])
    total_duration = total_samples / SAMPLE_RATE
    fs = get_feature_stats(feature_mode=mixdb.feature,
                           frame_size=mixdb.frame_size,
                           num_classes=mixdb.num_classes,
                           truth_mutex=mixdb.truth_mutex)

    logger.info(f'{"Mixtures":{desc_len}} {len(mixdb.mixtures)}')
    logger.info(f'{"Duration":{desc_len}} {seconds_to_hms(seconds=total_duration)}')
    logger.info(f'{"Targets":{desc_len}} {len(mixdb.targets)}')
    logger.info(f'{"Target augmentations":{desc_len}} {len(mixdb.target_augmentations)}')
    logger.info(f'{"Noise augmentations":{desc_len}} {len(mixdb.noise_augmentations)}')
    logger.info(f'{"Noises":{desc_len}} {len(mixdb.noises)}')
    logger.info(f'{"Feature":{desc_len}} {mixdb.feature}')
    logger.info(f'{"Feature shape":{desc_len}} {fs.stride} x {fs.num_bands} ({fs.stride * fs.num_bands} total params)')
    logger.info(f'{"Feature samples":{desc_len}} {fs.feature_samples} samples ({fs.feature_ms} ms)')
    logger.info(f'{"Feature step samples":{desc_len}} {fs.feature_step_samples} samples ({fs.feature_step_ms} ms)')
    logger.info(f'{"Feature overlap":{desc_len}} {fs.step / fs.stride} ({fs.feature_step_ms} ms)')
    logger.info(f'{"SNRs":{desc_len}} {mixdb.snrs}')
    logger.info(f'{"Random SNRs":{desc_len}} {mixdb.random_snrs}')
    logger.info(f'{"Classes":{desc_len}} {mixdb.num_classes}')
    logger.info(f'{"Truth mutex":{desc_len}} {mixdb.truth_mutex}')
    print_class_count(record=mixdb, length=desc_len, print_fn=logger.info)
    # TBD add class weight calculations here
    logger.info('')

    if list_targets:
        logger.info('Target details:')
        idx_len = int(np.ceil(np.log10(len(mixdb.targets))))
        for idx, target in enumerate(mixdb.targets):
            desc = f'  {idx:{idx_len}} Name'
            logger.info(f'{desc:{desc_len}} {target.name}')
            desc = f'  {idx:{idx_len}} Truth index'
            logger.info(f'{desc:{desc_len}} {get_truth_indices_for_target(target)}')
        logger.info('')

    if truth_index is not None:
        if 0 <= truth_index > mixdb.num_classes:
            raise SonusAIError(f'Given truth_index is outside valid range of 1-{mixdb.num_classes}')
        ids = get_mixids_from_truth_index(mixdb=mixdb, predicate=lambda x: x in [truth_index])[truth_index]
        logger.info(f'Mixtures with truth index {truth_index}: {ids}')
        logger.info('')

    if mixid == '*':
        mixid = list(range(len(mixdb.mixtures)))
    else:
        mixid = expand_range(mixid)

    if len(mixid) == 1:
        print_mixture_details(mixdb=mixdb,
                              mixid=mixid[0],
                              desc_len=desc_len,
                              print_fn=logger.info,
                              all_class_counts=True)
    else:
        if truth_f_shape is not None:
            logger.info(f'Calculating statistics from .h5 file truth for {len(mixid):,} mixtures'
                        f' ({consolidate_range(mixid)})')
            with h5py.File(datafile, 'r') as f:
                truth_f = np.array(f['truth_f'])

            snr_mean, snr_std, snr_db_mean, snr_db_std = calculate_snr_f_statistics(truth_f)

            logger.info('Truth')
            logger.info(f'  {"mean":^8s}  {"std":^8s}  {"db_mean":^8s}  {"db_std":^8s}')
            for c in range(len(snr_mean)):
                logger.info(f'  {snr_mean[c]:8.2f}  {snr_std[c]:8.2f}  {snr_db_mean[c]:8.2f}  {snr_db_std[c]:8.2f}')
        else:
            logger.info(f'Generating truth for statistics of {len(mixdb.mixtures)} mixtures ...')
            # TBD run genmix on small batches of mixtures and calculate running statistics


def main():
    import os

    import h5py
    from docopt import docopt

    import sonusai
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import load_mixdb
    from sonusai.utils import human_readable_size
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    mixdb_name = args['MIXDB']
    mixid = args['--mixid']
    truth_index = args['--truth_index']
    list_targets = args['--targets']

    if truth_index is not None:
        truth_index = int(truth_index)

    log_name = 'lsdb.log'
    create_file_handler(log_name)
    update_console_handler(False)
    initial_log_messages('lsdb')

    logger.info(f'Analyzing {mixdb_name} ({human_readable_size(os.path.getsize(mixdb_name), 1)})')

    mixdb = load_mixdb(name=mixdb_name)

    datafile = None
    if h5py.is_hdf5(mixdb_name):
        datafile = mixdb_name

    lsdb(mixdb=mixdb,
         mixid=mixid,
         datafile=datafile,
         truth_index=truth_index,
         list_targets=list_targets)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



import numpy as np


def read_mixture_data(filename: str) -> (np.ndarray, np.ndarray, np.ndarray):
    """Read mixture, target, and noise data from given HDF5 file and return them as a tuple."""
    import h5py

    if not filename:
        return None, None, None

    with h5py.File(name=filename, mode='r') as f:
        return np.array(f['/mixture']), np.array(f['/target']), np.array(f['/noise'])



from typing import Callable
from typing import Union

from sonusai.mixture import Mixture
from sonusai.mixture import MixtureDatabase


def print_mixture_details(mixdb: MixtureDatabase,
                          mixid: int = None,
                          desc_len: int = 1,
                          print_fn: Callable = print,
                          all_class_counts: bool = False) -> None:
    import numpy as np

    from sonusai import SonusAIError
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import get_feature_frames_in_mixture

    if mixid is not None:
        if 0 < mixid >= len(mixdb.mixtures):
            raise SonusAIError(f'Given mixid is outside valid range of 0:{len(mixdb.mixtures) - 1}.')

        print_fn(f'Mixture {mixid} details:')
        mixture = mixdb.mixtures[mixid]
        targets = [mixdb.targets[idx] for idx in mixture.target_file_index]
        target_augmentations = [mixdb.target_augmentations[idx] for idx in mixture.target_augmentation_index]
        noise = mixdb.noises[mixture.noise_file_index]
        print_fn(f'Target files:')
        for idx, target in enumerate(targets):
            print_fn(f'{"  Name":{desc_len}} {target.name}')
            print_fn(f'{"  Duration":{desc_len}} {target.duration}')
            print_fn(f'  Truth settings:')
            for truth_setting in target.truth_settings:
                print_fn(f'{"    Index":{desc_len}} {truth_setting.index}')
                print_fn(f'{"    Function":{desc_len}} {truth_setting.function}')
                print_fn(f'{"    Config":{desc_len}} {truth_setting.config}')
            print_fn(f'{"  Augmentation":{desc_len}} {target_augmentations[idx]}')
        print_fn(f'{"Samples":{desc_len}} {mixture.samples}')
        print_fn(f'{"Features":{desc_len}} {get_feature_frames_in_mixture(mixdb, mixid)}')
        print_fn(f'{"Noise":{desc_len}} {noise.name}')
        noise_offset_percent = int(np.round(100 * mixture.noise_offset / float(noise.duration * SAMPLE_RATE)))
        print_fn(f'{"Noise offset":{desc_len}} {mixture.noise_offset} samples ({noise_offset_percent}%)')
        print_fn(f'{"SNR":{desc_len}} {mixture.snr} dB')
        print_fn(f'{"Target gain":{desc_len}} {mixture.target_gain}')
        print_fn(f'{"Target SNR gain":{desc_len}} {mixture.target_snr_gain}')
        print_fn(f'{"Noise SNR gain":{desc_len}} {mixture.noise_snr_gain}')
        print_class_count(record=mixture,
                          length=desc_len,
                          print_fn=print_fn,
                          all_class_counts=all_class_counts)
        print_fn('')


def print_class_count(record: Union[MixtureDatabase, Mixture],
                      length: int,
                      print_fn: Callable = print,
                      all_class_counts: bool = False) -> None:
    import numpy as np

    print_fn(f'Class count:')
    idx_len = int(np.ceil(np.log10(len(record.class_count))))
    for idx, count in enumerate(record.class_count):
        if all_class_counts or count > 0:
            desc = f'  class {idx + 1:{idx_len}}'
            print_fn(f'{desc:{length}} {count}')



from typing import Any
from typing import Callable
from typing import Dict
from typing import List

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureID


def get_mixids_from_mixture_field_predicate(mixdb: MixtureDatabase,
                                            field: str,
                                            mixid: MixtureID = None,
                                            predicate: Callable[[Any], bool] = None) -> Dict[int, List[int]]:
    """
    Generate mixture IDs based on mixture field and predicate
    Return a dictionary where:
        - keys are the matching field values
        - values are lists of the mixids that match the criteria
    """
    from typing import Any

    from sonusai.mixture import convert_mixid_to_list

    mixid_out = convert_mixid_to_list(mixdb, mixid)

    if predicate is None:
        def predicate(_: Any) -> bool:
            return True

    criteria = set()
    for i, x in enumerate(mixdb.mixtures):
        if i in mixid_out:
            value = getattr(x, field)
            if isinstance(value, list):
                for v in value:
                    if predicate(v):
                        criteria.add(v)
            elif predicate(value):
                criteria.add(value)
    criteria = sorted(list(criteria))

    result = dict()
    for criterion in criteria:
        result[criterion] = list()
        for i, x in enumerate(mixdb.mixtures):
            if i in mixid_out:
                value = getattr(x, field)
                if isinstance(value, list):
                    for v in value:
                        if v == criterion:
                            result[criterion].append(i)
                elif value == criterion:
                    result[criterion].append(i)

    return result


def get_mixids_from_truth_settings_field_predicate(mixdb: MixtureDatabase,
                                                   field: str,
                                                   mixid: MixtureID = None,
                                                   predicate: Callable[[Any], bool] = None) -> Dict[int, List[int]]:
    """
    Generate mixture IDs based on target truth_settings field and predicate
    Return a dictionary where:
        - keys are the matching field values
        - values are lists of the mixids that match the criteria
    """
    from typing import Any

    from sonusai.mixture import convert_mixid_to_list

    mixid_out = convert_mixid_to_list(mixdb, mixid)

    # Get all field values
    values = get_all_truth_settings_values_from_field(mixdb, field)

    if predicate is None:
        def predicate(_: Any) -> bool:
            return True

    # Get only values of interest
    values = [value for value in values if predicate(value)]

    result = dict()
    for value in values:
        # Get a list of targets for each field value
        indices = list()
        for i, target in enumerate(mixdb.targets):
            for truth_setting in target.truth_settings:
                if value in getattr(truth_setting, field):
                    indices.append(i)
        indices = sorted(list(set(indices)))

        mixids = list()
        for index in indices:
            mixids.extend([i for i, x in enumerate(mixdb.mixtures) if
                           index in x.target_file_index and i in mixid_out])

        mixids = sorted(list(set(mixids)))
        if mixids:
            result[value] = mixids

    return result


def get_all_truth_settings_values_from_field(mixdb: MixtureDatabase, field: str) -> list:
    """
    Generate a list of all values corresponding to the given field in truth_settings
    """
    result = list()
    for target in mixdb.targets:
        for truth_setting in target.truth_settings:
            value = getattr(truth_setting, field)
            if isinstance(value, str):
                value = [value]
            result.extend(value)

    return sorted(list(set(result)))


def get_mixids_from_noise(mixdb: MixtureDatabase,
                          mixid: MixtureID = None,
                          predicate: Callable[[Any], bool] = None) -> Dict[int, List[int]]:
    """
    Generate mixids based on noise index predicate
    Return a dictionary where:
        - keys are the noise indices
        - values are lists of the mixids that match the noise index
    """
    return get_mixids_from_mixture_field_predicate(mixdb=mixdb,
                                                   mixid=mixid,
                                                   field='noise_file_index',
                                                   predicate=predicate)


def get_mixids_from_noise_augmentation(mixdb: MixtureDatabase,
                                       mixid: MixtureID = None,
                                       predicate: Callable[[Any], bool] = None) -> Dict[int, List[int]]:
    """
    Generate mixids based on a noise augmentation index predicate
    Return a dictionary where:
        - keys are the noise augmentation indices
        - values are lists of the mixids that match the noise augmentation index
    """
    return get_mixids_from_mixture_field_predicate(mixdb=mixdb,
                                                   mixid=mixid,
                                                   field='noise_augmentation_index',
                                                   predicate=predicate)


def get_mixids_from_target(mixdb: MixtureDatabase,
                           mixid: MixtureID = None,
                           predicate: Callable[[Any], bool] = None) -> Dict[int, List[int]]:
    """
    Generate mixids based on a target index predicate
    Return a dictionary where:
        - keys are the target indices
        - values are lists of the mixids that match the target index
    """
    return get_mixids_from_mixture_field_predicate(mixdb=mixdb,
                                                   mixid=mixid,
                                                   field='target_file_index',
                                                   predicate=predicate)


def get_mixids_from_target_augmentation(mixdb: MixtureDatabase,
                                        mixid: MixtureID = None,
                                        predicate: Callable[[Any], bool] = None) -> Dict[int, List[int]]:
    """
    Generate mixids based on a target augmentation index predicate
    Return a dictionary where:
        - keys are the target augmentation indices
        - values are lists of the mixids that match the target augmentation index
    """
    return get_mixids_from_mixture_field_predicate(mixdb=mixdb,
                                                   mixid=mixid,
                                                   field='target_augmentation_index',
                                                   predicate=predicate)


def get_mixids_from_snr(mixdb: MixtureDatabase,
                        mixid: MixtureID = None,
                        predicate: Callable[[Any], bool] = None) -> Dict[float, List[int]]:
    """
    Generate mixids based on an SNR predicate
    Return a dictionary where:
        - keys are the SNRs
        - values are lists of the mixids that match the SNR
    """
    from typing import Any

    from sonusai.mixture import convert_mixid_to_list

    mixid_out = convert_mixid_to_list(mixdb, mixid)

    # Get all the SNRs
    snrs = mixdb.snrs

    if predicate is None:
        def predicate(_: Any) -> bool:
            return True

    # Get only the SNRs of interest (filter on predicate)
    snrs = [snr for snr in snrs if predicate(snr)]

    result = dict()
    for snr in snrs:
        # Get a list of mixids for each SNR
        result[snr] = sorted(
            [i for i, x in enumerate(mixdb.mixtures) if x.snr == snr and i in mixid_out and not x.random_snr])

    return result


def get_mixids_from_truth_index(mixdb: MixtureDatabase,
                                mixid: MixtureID = None,
                                predicate: Callable[[Any], bool] = None) -> Dict[int, List[int]]:
    """
    Generate mixids based on a truth index predicate
    Return a dictionary where:
        - keys are the truth indices
        - values are lists of the mixids that match the truth index
    """
    return get_mixids_from_truth_settings_field_predicate(mixdb=mixdb,
                                                          mixid=mixid,
                                                          field='index',
                                                          predicate=predicate)


def get_mixids_from_truth_function(mixdb: MixtureDatabase,
                                   mixid: MixtureID = None,
                                   predicate: Callable[[Any], bool] = None) -> Dict[int, List[int]]:
    """
    Generate mixids based on a truth function predicate
    Return a dictionary where:
        - keys are the truth functions
        - values are lists of the mixids that match the truth function
    """
    return get_mixids_from_truth_settings_field_predicate(mixdb=mixdb,
                                                          mixid=mixid,
                                                          field='function',
                                                          predicate=predicate)



import numpy as np
from tensorflow.keras.utils import Sequence

from sonusai.mixture import MixtureID


class DataGenerator(Sequence):
    """Generates data for Keras"""

    def __init__(self,
                 filename: str,
                 mixid: MixtureID,
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False):
        """Initialization"""
        import h5py

        from sonusai import SonusAIError
        from sonusai.mixture import Segment
        from sonusai.mixture import convert_mixid_to_list
        from sonusai.mixture import get_feature_frames_in_mixture
        from sonusai.mixture import mixdb_from_json

        self.filename = filename
        self.mixid = mixid
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle

        try:
            self.file = h5py.File(self.filename, 'r')
            self.mixdb = mixdb_from_json(self.file.attrs['mixdb'])
            self.stride = self.file['feature'].shape[1]
            self.num_bands = self.file['feature'].shape[2]
            self.num_classes = self.file['truth_f'].shape[1]
        except Exception as e:
            raise SonusAIError(f'Error: {e}')

        self.mixid = convert_mixid_to_list(self.mixdb, self.mixid)

        self.file_frame_segments = dict()
        for m in self.mixid:
            self.file_frame_segments[m] = Segment(self.mixdb.mixtures[m].o_frame_offset,
                                                  get_feature_frames_in_mixture(self.mixdb, m))

        self.mixtures = None
        self.mixture_frame_segments = None
        self.batch_frame_segments = None
        self.total_batches = None
        self.frames_per_batch = None

        self._initialize_mixtures()

    def __len__(self) -> int:
        """Denotes the number of batches per epoch"""
        return self.total_batches

    def __getitem__(self, batch_index: int) -> (np.ndarray, np.ndarray):
        """Get one batch of data"""

        from sonusai.utils import reshape_inputs

        feature = np.empty((self.frames_per_batch, self.stride, self.num_bands), dtype=np.single)
        truth = np.empty((self.frames_per_batch, self.num_classes), dtype=np.single)
        start = 0
        for segment in self.batch_frame_segments[batch_index]:
            length = segment.length
            feature[start:start + length] = self.file['feature'][segment.get_slice()]
            truth[start:start + length] = self.file['truth_f'][segment.get_slice()]
            start += length

        feature, truth, _, _, _, _ = reshape_inputs(feature=feature,
                                                    truth=truth,
                                                    batch_size=self.batch_size,
                                                    timesteps=self.timesteps,
                                                    flatten=self.flatten,
                                                    add1ch=self.add1ch)
        return feature, truth

    def on_epoch_end(self) -> None:
        """Modification of dataset between epochs"""
        import random

        if self.shuffle:
            random.shuffle(self.mixid)
            self.initialize_mixtures()

    def _initialize_mixtures(self) -> None:
        from copy import copy

        from sonusai import SonusAIError
        from sonusai.mixture import Segment
        from sonusai.mixture import get_mixtures_from_mixid

        self.mixtures = get_mixtures_from_mixid(self.mixdb, self.mixid)
        self.mixture_frame_segments = [self.file_frame_segments[m] for m in self.mixid]

        frames = sum([sub.samples for sub in self.mixtures]) // self.mixdb.feature_step_samples
        self.frames_per_batch = self.batch_size if self.timesteps == 0 else self.batch_size * self.timesteps
        self.total_batches = frames // self.frames_per_batch

        if self.total_batches == 0:
            raise SonusAIError(
                f'Error: dataset only contains {frames} frames which is not enough to fill a batch size of '
                f'{self.frames_per_batch}. Either provide more data or decrease the batch size')

        # Compute mixid and offset for dataset
        # offsets are needed because mixtures are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence.
        cumulative_frames = 0
        start_mixture_index = 0
        offset = 0
        index_map = list()
        for m in range(len(self.mixture_frame_segments)):
            current_frames = self.mixture_frame_segments[m].length
            cumulative_frames += current_frames
            while cumulative_frames >= self.frames_per_batch:
                extra_frames = cumulative_frames - self.frames_per_batch
                index_map.append({
                    'mixid':  list(range(start_mixture_index, m + 1)),
                    'offset': offset,
                    'extra':  extra_frames
                })
                if extra_frames == 0:
                    start_mixture_index = m + 1
                    offset = 0
                else:
                    start_mixture_index = m
                    offset = current_frames - extra_frames
                cumulative_frames = extra_frames

        self.batch_frame_segments = list()
        for item in index_map:
            slices = list()
            for m in item['mixid']:
                slices.append(copy(self.mixture_frame_segments[m]))
            slices[0].trim_start(item['offset'])
            slices[-1].trim_length(item['extra'])
            consolidated_segments = list()
            start = slices[0].start
            length = slices[0].length
            for i in range(1, len(slices)):
                if slices[i].start != start + length:
                    consolidated_segments.append(Segment(start, length))
                    start = slices[i].start
                    length = slices[i].length
                else:
                    length += slices[i].length
            consolidated_segments.append(Segment(start, length))

            self.batch_frame_segments.append(consolidated_segments)



from dataclasses import dataclass
from typing import Union

import numpy as np
from pyaaware import Predict

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureID


def get_ft_from_file(filename: str, mixid: MixtureID = ':') -> (np.ndarray, np.ndarray):
    """Get feature/truth frames from H5 file for given mixture ID's"""
    import h5py

    from sonusai import SonusAIError
    from sonusai.mixture import convert_mixid_to_list
    from sonusai.mixture import get_file_frame_segments
    from sonusai.mixture import mixdb_from_json
    from sonusai.mixture import get_slices

    try:
        with h5py.File(filename, 'r') as f:
            mixdb = mixdb_from_json(f.attrs['mixdb'])

            in_frames, stride, num_bands = f['/feature'].shape

            if f['/truth_f'].shape[0] != in_frames:
                raise SonusAIError('feature and truth_f have different number of frames')

            num_classes = f['/truth_f'].shape[1]

            _mixid = convert_mixid_to_list(mixdb, mixid)
            file_frame_segments = get_file_frame_segments(mixdb, _mixid)
            out_frames = sum([file_frame_segments[m].length for m in _mixid])

            feature = np.empty((out_frames, stride, num_bands), dtype=np.single)
            truth = np.empty((out_frames, num_classes), dtype=np.single)

            start = 0
            for m in _mixid:
                in_slice, out_slice = get_slices(file_frame_segments[m], start)
                feature[out_slice] = f['/feature'][in_slice]
                truth[out_slice] = f['/truth_f'][in_slice]
                start += file_frame_segments[m].length

            return feature, truth

    except Exception as e:
        raise SonusAIError(f'Error: {e}')


def get_feature_from_audio(audio: np.ndarray, model: Predict) -> np.ndarray:
    import h5py

    import sonusai
    from sonusai.mixture import get_pad_length
    from sonusai.mixture import Mixture

    fs = get_feature_stats(feature_mode=model.feature,
                           frame_size=sonusai.mixture.DEFAULT_FRAME_SIZE,
                           num_classes=1,
                           truth_mutex=model.mutex)
    audio = np.pad(array=audio,
                   pad_width=(0, get_pad_length(len(audio), fs.feature_step_samples)),
                   mode='constant',
                   constant_values=0)
    mixdb = MixtureDatabase(feature=model.feature,
                            mixtures=[Mixture(samples=len(audio))],
                            feature_samples=fs.feature_samples,
                            feature_step_samples=fs.feature_step_samples,
                            frame_size=sonusai.mixture.DEFAULT_FRAME_SIZE,
                            num_classes=1,
                            truth_mutex=model.mutex,
                            truth_reduction_function='max')
    feature, _ = get_feature_and_truth_f(mixdb=mixdb, mixid=0, audio=audio)
    return feature


def read_feature_data(filename: str) -> (MixtureDatabase, np.ndarray, np.ndarray, np.ndarray):
    """Read mixdb, feature, truth_f, and segsnr data from given HDF5 file and return them as a tuple."""
    import h5py

    from sonusai.mixture import mixdb_from_json

    with h5py.File(name=filename, mode='r') as f:
        return (mixdb_from_json(f.attrs['mixdb']),
                np.array(f['/feature']),
                np.array(f['/truth_f']),
                np.array(f['/segsnr']))


def get_feature_and_truth_f(mixdb: MixtureDatabase,
                            mixid: int,
                            audio: np.ndarray,
                            truth_t: Union[np.ndarray, None] = None) -> (np.ndarray, np.ndarray):
    from pyaaware import FeatureGenerator
    from pyaaware import ForwardTransform

    from sonusai import SonusAIError
    from sonusai.mixture import get_feature_frames_in_mixture
    from sonusai.mixture import get_transform_frames_in_mixture
    from sonusai.mixture import truth_reduction
    from sonusai.utils import int16_to_float

    if len(audio) != mixdb.mixtures[mixid].samples:
        raise SonusAIError(f'Wrong number of samples in audio')

    fft = ForwardTransform(N=mixdb.frame_size * 4,
                           R=mixdb.frame_size)

    fg = FeatureGenerator(frame_size=mixdb.frame_size,
                          feature_mode=mixdb.feature,
                          num_classes=mixdb.num_classes,
                          truth_mutex=mixdb.truth_mutex)

    transform_frames = get_transform_frames_in_mixture(mixdb, mixid)
    feature_frames = get_feature_frames_in_mixture(mixdb, mixid)

    if truth_t is None:
        truth_t = np.zeros((mixdb.mixtures[mixid].samples, mixdb.num_classes), dtype=np.single)

    feature = np.empty((feature_frames, fg.stride, fg.num_bands), dtype=np.single)
    truth_f = np.empty((feature_frames, mixdb.num_classes), dtype=np.single)

    feature_frame = 0
    for transform_frame in range(transform_frames):
        indices = slice(transform_frame * mixdb.frame_size,
                        (transform_frame + 1) * mixdb.frame_size)
        fd = fft.execute(int16_to_float(audio[indices]))

        fg.execute(fd, truth_reduction(truth_t[indices], mixdb.truth_reduction_function))

        if fg.eof():
            feature[feature_frame] = fg.feature()
            truth_f[feature_frame] = fg.truth()
            feature_frame += 1

    return feature, truth_f


@dataclass(frozen=True)
class FeatureStats:
    feature_ms: float
    feature_samples: int
    feature_step_ms: float
    feature_step_samples: int
    num_bands: int
    stride: int
    step: int
    decimation: int


def get_feature_stats(feature_mode: str,
                      frame_size: int,
                      num_classes: int,
                      truth_mutex: bool) -> FeatureStats:
    import h5py
    from pyaaware import FeatureGenerator

    import sonusai

    fg = FeatureGenerator(feature_mode=feature_mode,
                          frame_size=frame_size,
                          num_classes=num_classes,
                          truth_mutex=truth_mutex)

    transform_frame_ms = float(frame_size) / float(sonusai.mixture.SAMPLE_RATE / 1000)

    return FeatureStats(feature_ms=transform_frame_ms * fg.decimation * fg.stride,
                        feature_samples=frame_size * fg.decimation * fg.stride,
                        feature_step_ms=transform_frame_ms * fg.decimation * fg.step,
                        feature_step_samples=frame_size * fg.decimation * fg.step,
                        num_bands=fg.num_bands,
                        stride=fg.stride,
                        step=fg.step,
                        decimation=fg.decimation)



def get_label_names(num_labels: int, file: str = None) -> list:
    """Return label names in a list. Read from CSV file, if provided."""
    import csv

    if file is None:
        return [f'Class {val + 1}' for val in range(num_labels)]

    label_names = [''] * num_labels
    with open(file) as f:
        reader = csv.DictReader(f)
        if 'index' not in reader.fieldnames or 'display_name' not in reader.fieldnames:
            raise Exception(f'Missing required fields in labels CSV.')

        for row in reader:
            index = int(row['index']) - 1
            if index >= num_labels:
                raise Exception(f'The number of given label names does not match the number of labels.')
            label_names[index] = row['display_name']

    return label_names



def dataclass_from_dict(klass, dikt):
    """Convert dictionary to dataclass."""
    try:
        fieldtypes = klass.__annotations__
        return klass(**{f: dataclass_from_dict(fieldtypes[f], dikt[f]) for f in dikt})
    except AttributeError:
        if isinstance(dikt, (tuple, list)):
            return [dataclass_from_dict(klass.__args__[0], f) for f in dikt]
        return dikt



"""sonusai genft

usage: genft [-hvs] (-d MIXDB) [-i MIXID] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Mixture database JSON file.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate and analyze. [default: *].
    -o OUTPUT, --output OUTPUT      Output HDF5 file.
    -s, --segsnr                    Save segsnr. [default: False].

Generate a SonusAI feature/truth file from a SonusAI mixture database.

Inputs:
    MIXDB       A SonusAI mixture database JSON file.

Outputs:
    OUTPUT.h5   A SonusAI feature HDF5 file. Contains:
                    dataset:    feature
                    dataset:    truth_f
                    dataset:    segsnr (optional)
                    attribute:  mixdb
    genft.log

"""
import multiprocessing as mp

import numpy as np
from tqdm import tqdm

from sonusai import logger
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import MixtureID

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = dict()


# noinspection PyGlobalUndefined

def mp_init(feature_: mp.Array,
            truth_f_: mp.Array,
            segsnr_: mp.Array) -> None:
    global mp_feature
    global mp_truth_f
    global mp_segsnr

    mp_feature = feature_
    mp_truth_f = truth_f_
    mp_segsnr = segsnr_


def init(mixdb: MixtureDatabase,
         mixid: MixtureID,
         logging: bool = True) -> (MixtureDatabase, int, int, int):
    from sonusai.mixture import new_mixdb_from_mixid

    mixdb_out = new_mixdb_from_mixid(mixdb=mixdb, mixid=mixid)

    total_samples = sum([sub.samples for sub in mixdb_out.mixtures])
    total_transform_frames = total_samples // mixdb_out.frame_size
    total_feature_frames = total_samples // mixdb_out.feature_step_samples

    if logging:
        logger.info('')
        logger.info(f'Found {len(mixdb_out.mixtures):,} mixtures to process')
        logger.info(f'{total_samples:,} samples, '
                    f'{total_transform_frames:,} transform frames, '
                    f'{total_feature_frames:,} feature frames')

    return mixdb_out, total_samples, total_transform_frames, total_feature_frames


def genft(mixdb: MixtureDatabase,
          mixid: MixtureID,
          compute_segsnr: bool = False,
          logging: bool = True,
          show_progress: bool = False,
          progress: tqdm = None,
          initial_i_sample_offset: int = 0,
          initial_i_frame_offset: int = 0,
          initial_o_frame_offset: int = 0) -> (np.ndarray,
                                               np.ndarray,
                                               np.ndarray,
                                               MixtureDatabase):
    import ctypes

    from pyaaware import FeatureGenerator
    from tqdm import tqdm

    import sonusai
    from sonusai.mixture import build_noise_audios
    from sonusai.mixture import build_target_audios
    from sonusai.mixture import get_total_class_count
    from sonusai.mixture import set_mixture_offsets
    from sonusai.utils import human_readable_size
    from sonusai.utils import p_tqdm_map
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import to_numpy_array

    mixdb_out, total_samples, total_transform_frames, total_feature_frames = init(mixdb=mixdb, mixid=mixid,
                                                                                  logging=logging)

    MP_DICT['mixdb'] = mixdb_out
    MP_DICT['total_feature_frames'] = total_feature_frames
    MP_DICT['compute_segsnr'] = compute_segsnr
    MP_DICT['target_audios'] = build_target_audios(mixdb=mixdb_out, show_progress=show_progress)
    MP_DICT['noise_audios'] = build_noise_audios(mixdb=mixdb_out, show_progress=show_progress)

    fg = FeatureGenerator(frame_size=mixdb_out.frame_size,
                          feature_mode=mixdb_out.feature,
                          num_classes=mixdb_out.num_classes,
                          truth_mutex=mixdb_out.truth_mutex)
    mp_feature = mp.Array(ctypes.c_float, total_feature_frames * fg.stride * fg.num_bands)
    mp_truth_f = mp.Array(ctypes.c_float, total_feature_frames * fg.num_classes)
    if compute_segsnr:
        mp_segsnr = mp.Array(ctypes.c_float, total_transform_frames)
    else:
        mp_segsnr = mp.Array(ctypes.c_float, 0)

    feature = to_numpy_array(mp_feature, dtype=np.single).reshape((total_feature_frames, fg.stride, fg.num_bands))
    truth_f = to_numpy_array(mp_truth_f, dtype=np.single).reshape((total_feature_frames, fg.num_classes))
    if compute_segsnr:
        segsnr = to_numpy_array(mp_segsnr, dtype=np.single)
    else:
        segsnr = np.empty(0, dtype=np.single)

    # First pass to get offsets
    set_mixture_offsets(mixdb=mixdb_out,
                        initial_i_sample_offset=initial_i_sample_offset,
                        initial_i_frame_offset=initial_i_frame_offset,
                        initial_o_frame_offset=initial_o_frame_offset)

    # Second pass to get feature and truth_f
    progress_needs_close = False
    if progress is None:
        progress = tqdm(total=len(mixdb_out.mixtures), desc='genft', disable=not show_progress)
        progress_needs_close = True

    p_tqdm_map(_process_mixture,
               list(range(len(mixdb_out.mixtures))),
               progress=progress,
               initializer=mp_init,
               initargs=(mp_feature, mp_truth_f, mp_segsnr))

    if progress_needs_close:
        progress.close()

    mixdb_out.class_count = get_total_class_count(mixdb_out)

    duration = total_samples / sonusai.mixture.SAMPLE_RATE
    if logging:
        logger.info('')
        logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
        logger.info(f'feature:  {human_readable_size(feature.nbytes, 1)}')
        logger.info(f'truth_f:  {human_readable_size(truth_f.nbytes, 1)}')
        if compute_segsnr:
            logger.info(f'segsnr:   {human_readable_size(segsnr.nbytes, 1)}')

    return feature, truth_f, segsnr, mixdb_out


def _process_mixture(mixid: int) -> None:
    from pyaaware import FeatureGenerator

    from sonusai.mixture import get_audio_and_truth_t
    from sonusai.mixture import get_feature_and_truth_f
    from sonusai.mixture import get_feature_frames_in_mixture
    from sonusai.mixture import get_transform_frames_in_mixture
    from sonusai.utils import to_numpy_array

    mixdb: MixtureDatabase = MP_DICT['mixdb']
    total_feature_frames = MP_DICT['total_feature_frames']
    target_audios = MP_DICT['target_audios']
    noise_audios = MP_DICT['noise_audios']
    compute_segsnr = MP_DICT['compute_segsnr']

    fg = FeatureGenerator(frame_size=mixdb.frame_size,
                          feature_mode=mixdb.feature,
                          num_classes=mixdb.num_classes,
                          truth_mutex=mixdb.truth_mutex)

    feature = to_numpy_array(mp_feature, dtype=np.single)
    feature = feature.reshape((total_feature_frames, fg.stride, fg.num_bands))

    truth_f = to_numpy_array(mp_truth_f, dtype=np.single)
    truth_f = truth_f.reshape((total_feature_frames, mixdb.num_classes))

    segsnr = to_numpy_array(mp_segsnr, dtype=np.single)

    record = mixdb.mixtures[mixid]
    first_i_frame_offset = mixdb.mixtures[0].i_frame_offset
    transform_frames = get_transform_frames_in_mixture(mixdb, mixid)
    feature_frames = get_feature_frames_in_mixture(mixdb, mixid)

    frame_indices = slice(record.i_frame_offset - first_i_frame_offset,
                          record.i_frame_offset - first_i_frame_offset + transform_frames)

    (mixture_td,
     truth_t,
     _,
     _,
     segsnr[frame_indices]) = get_audio_and_truth_t(mixdb=mixdb,
                                                    mixture=record,
                                                    raw_target_audios=target_audios,
                                                    raw_noise_audios=noise_audios,
                                                    compute_truth=True,
                                                    compute_segsnr=compute_segsnr,
                                                    frame_based_segsnr=True)

    first_o_frame_offset = mixdb.mixtures[0].o_frame_offset
    frame_indices = slice(record.o_frame_offset - first_o_frame_offset,
                          record.o_frame_offset - first_o_frame_offset + feature_frames)
    (feature[frame_indices],
     truth_f[frame_indices]) = get_feature_and_truth_f(mixdb=mixdb,
                                                       mixid=mixid,
                                                       audio=mixture_td,
                                                       truth_t=truth_t)


def main():
    import time
    from os.path import splitext

    import h5py
    from docopt import docopt
    from pyaaware import FeatureGenerator
    from tqdm import tqdm

    import sonusai
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import check_audio_files_exist
    from sonusai.mixture import get_feature_frames_in_mixture
    from sonusai.mixture import get_samples_in_mixture
    from sonusai.mixture import get_transform_frames_in_mixture
    from sonusai.mixture import load_mixdb
    from sonusai.utils import expand_range
    from sonusai.utils import grouper
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixdb_name = args['--mixdb']
    mixid = args['--mixid']
    output_name = args['--output']
    compute_segsnr = args['--segsnr']

    if not output_name:
        output_name = splitext(mixdb_name)[0] + '.h5'

    start_time = time.monotonic()

    log_name = 'genft.log'
    create_file_handler(log_name)
    update_console_handler(verbose)
    initial_log_messages('genft')

    logger.info(f'\nLoad mixture database from {mixdb_name}')
    mixdb = load_mixdb(name=mixdb_name)
    if mixid == '*':
        mixid = list(range(len(mixdb.mixtures)))
    else:
        mixid = expand_range(mixid)

    mixdb_out, total_samples, total_transform_frames, total_feature_frames = init(mixdb=mixdb,
                                                                                  mixid=mixid,
                                                                                  logging=True)

    check_audio_files_exist(mixdb)

    fg = FeatureGenerator(frame_size=mixdb_out.frame_size,
                          feature_mode=mixdb_out.feature,
                          num_classes=mixdb_out.num_classes,
                          truth_mutex=mixdb_out.truth_mutex)

    with h5py.File(output_name, 'w') as f:
        f.create_dataset(name='feature',
                         shape=(total_feature_frames, fg.stride, fg.num_bands),
                         dtype=np.single)
        f.create_dataset(name='truth_f',
                         shape=(total_feature_frames, fg.num_classes),
                         dtype=np.single)
        if compute_segsnr:
            f.create_dataset(name='segsnr',
                             shape=(total_transform_frames,),
                             dtype=np.single)

    chunk_size = 100
    progress = tqdm(total=len(mixid), desc='genft')
    mixid = grouper(range(len(mixdb_out.mixtures)), chunk_size)
    mixdb_out.class_count = [0] * mixdb_out.num_classes

    i_sample_offset = 0
    i_frame_offset = 0
    o_frame_offset = 0
    for m in mixid:
        feature, truth_f, segsnr, mixdb_tmp = genft(mixdb=mixdb_out,
                                                    mixid=m,
                                                    compute_segsnr=compute_segsnr,
                                                    logging=False,
                                                    progress=progress,
                                                    initial_i_sample_offset=i_sample_offset,
                                                    initial_i_frame_offset=i_frame_offset,
                                                    initial_o_frame_offset=o_frame_offset)
        o_frames = feature.shape[0]
        if o_frames != truth_f.shape[0]:
            logger.exception(f'truth_f frames does not match feature frames: {truth_f.shape[0]} != {o_frames}')
            raise SystemExit(1)

        with h5py.File(output_name, 'a') as f:
            indices = slice(o_frame_offset, o_frame_offset + o_frames)

            feature_dataset = f['feature']
            feature_dataset[indices] = feature

            truth_dataset = f['truth_f']
            truth_dataset[indices] = truth_f

            if compute_segsnr:
                segsnr_dataset = f['segsnr']
                segsnr_dataset[i_frame_offset:i_frame_offset + segsnr.shape[0]] = segsnr

        for idx, val in enumerate(m):
            mixdb_out.mixtures[val] = mixdb_tmp.mixtures[idx]
        for idx in range(mixdb_out.num_classes):
            mixdb_out.class_count[idx] += mixdb_tmp.class_count[idx]

        i_sample_offset = mixdb_out.mixtures[m[-1]].i_sample_offset
        i_sample_offset += get_samples_in_mixture(mixdb_out, m[-1])
        i_frame_offset = mixdb_out.mixtures[m[-1]].i_frame_offset
        i_frame_offset += get_transform_frames_in_mixture(mixdb_out, m[-1])
        o_frame_offset = mixdb_out.mixtures[m[-1]].o_frame_offset
        o_frame_offset += get_feature_frames_in_mixture(mixdb_out, m[-1])

    with h5py.File(output_name, 'a') as f:
        f.attrs['mixdb'] = mixdb_out.to_json()

    progress.close()

    logger.info(f'Wrote {output_name}')
    duration = total_samples / sonusai.mixture.SAMPLE_RATE
    logger.info('')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'feature:  {human_readable_size(o_frame_offset * fg.stride * fg.num_bands * 4, 1)}')
    logger.info(f'truth_f:  {human_readable_size(o_frame_offset * fg.num_classes * 4, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(i_frame_offset * 4, 1)}')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from typing import List

from sonusai.mixture.constants import DEFAULT_FRAME_SIZE


def calculate_input_shape(feature: str,
                          flatten: bool = False,
                          timesteps: int = 0,
                          add1ch: bool = False,
                          frame_size=DEFAULT_FRAME_SIZE) -> List[int]:
    """
    Calculate input shape given feature and user-specified reshape parameters.

    Inputs:
        feature:     String defining the Aaware feature used in SonusAI, typically  mixdb.feature.
        flatten:     If true, flatten the 2D spectrogram from SxB to S*B.
        timesteps:   Pre-pend timesteps dimension if non-zero, size = timesteps.
        add1ch:      Append channel dimension of size 1, (channel last).
        frame_size:  The default SonusAI frame size should always be used for now.
    """
    from pyaaware import FeatureGenerator

    # num_classes is irrelevant, set to 2
    fg = FeatureGenerator(frame_size=frame_size,
                          feature_mode=feature,
                          num_classes=2)

    if flatten:
        in_shape = [fg.stride * fg.num_bands]
    else:
        in_shape = [fg.stride, fg.num_bands]

    if timesteps > 0:
        in_shape.insert(0, timesteps)

    if add1ch:
        in_shape.append(1)

    return in_shape



"""sonusai predict

usage: predict [-hv] (-m MODEL) (-i INPUT) [-e EXPECT] [-o OUTPUT]

options:
   -h, --help
   -v, --verbose                Be verbose.
   -m MODEL, --model MODEL      Trained model ONNX file.
   -i INPUT, --input INPUT      Input feature WAV or HDF5 file.
   -e EXPECT, --expect EXPECT   Optional expected results HDF5 file.
   -o OUTPUT, --output OUTPUT   Output HDF5 file.

Run prediction on a trained model using SonusAI genft data and optionally
compare the results to expected results.

Inputs:
    MODEL   A SonusAI trained ONNX model file.
    INPUT   A WAV file; or an HDF5 file containing:
                dataset:    feature
    EXPECT  An HDF5 file containing:
                dataset:    predict

Outputs:
    OUTPUT  An HDF5 file containing:
                dataset:    predict
    predict.log

"""

from sonusai import logger


def main():
    import time
    from os.path import exists
    from os.path import splitext

    import h5py
    import numpy as np
    from docopt import docopt
    from pyaaware import Predict

    import sonusai
    from sonusai import SonusAIError
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import get_feature_from_audio
    from sonusai.mixture import read_audio
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    model_name = args['--model']
    input_name = args['--input']
    expect_name = args['--expect']
    output_name = args['--output']

    start_time = time.monotonic()

    log_name = 'predict.log'
    create_file_handler(log_name)
    update_console_handler(verbose)
    initial_log_messages('predict')

    logger.info('')
    logger.info(f'Model:  {model_name}')
    logger.info(f'Input:  {input_name}')
    logger.info(f'Expect: {expect_name}')
    logger.info(f'Output: {output_name}')
    logger.info('')

    if not output_name and not expect_name:
        raise SonusAIError('Must specify EXPECT or OUTPUT (or both).')

    model = Predict(model_name)
    logger.debug(f'Model feature name {model.feature}')
    logger.debug(f'Model input shape  {model.input_shape}')
    logger.debug(f'Model output shape {model.output_shape}')

    if not exists(input_name):
        raise SonusAIError(f'{input_name} does not exist')

    ext = splitext(input_name)[1]

    if ext == '.wav':
        audio = read_audio(input_name)
        feature = get_feature_from_audio(audio=audio, model=model)
    elif ext == '.h5':
        with h5py.File(name=input_name, mode='r') as f:
            feature = np.array(f['/feature'])
    else:
        raise SonusAIError(f'Unknown file type for {input_name}')

    logger.debug(f'Input shape        {feature.shape}')

    predict = model.execute(feature)

    if expect_name:
        with h5py.File(name=expect_name, mode='r') as f:
            expected = np.array(f['/predict'])
            logger.debug(f'Expect shape {expected.shape}')
            if expected.shape != predict.shape:
                raise SonusAIError('Expect shape does not match input shape')

            max_error = np.amax(np.abs(predict - expected))
            logger.info(f'Maximum error = {max_error}')

    if output_name:
        with h5py.File(name=output_name, mode='w') as f:
            f.create_dataset(name='predict', data=predict)
            logger.info(f'Wrote {output_name}')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



"""sonusai genmix

usage: genmix [-hvts] (-d MIXDB) [-i MIXID] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Mixture database JSON file.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate and analyze. [default: *].
    -o OUTPUT, --output OUTPUT      Output HDF5 file.
    -t, --truth                     Save truth_t. [default: False].
    -s, --segsnr                    Save segsnr. [default: False].

Generate a SonusAI mixture file from a SonusAI mixture database.

Inputs:
    MIXDB       A SonusAI mixture database JSON file.

Outputs:
    OUTPUT.h5   A SonusAI mixture HDF5 file. Contains:
                    dataset:    mixture
                    dataset:    truth_t (optional)
                    dataset:    target
                    dataset:    noise
                    dataset:    segsnr (optional)
                    attribute:  mixdb
    genmix.log
"""
import multiprocessing as mp

import numpy as np
from tqdm import tqdm

from sonusai import logger
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import MixtureID

# NOTE: multiprocessing dictionary is required for run-time performance; using 'partial' is much slower.
MP_DICT = dict()


# noinspection PyGlobalUndefined

def mp_init(mixture_: mp.Array,
            truth_t_: mp.Array,
            target_: mp.Array,
            noise_: mp.Array,
            segsnr_: mp.Array) -> None:
    global mp_mixture
    global mp_truth_t
    global mp_target
    global mp_noise
    global mp_segsnr

    mp_mixture = mixture_
    mp_truth_t = truth_t_
    mp_target = target_
    mp_noise = noise_
    mp_segsnr = segsnr_


def init(mixdb: MixtureDatabase,
         mixid: MixtureID,
         logging: bool = True) -> (MixtureDatabase, int):
    from sonusai.mixture import new_mixdb_from_mixid

    mixdb_out = new_mixdb_from_mixid(mixdb=mixdb, mixid=mixid)

    total_samples = sum([sub.samples for sub in mixdb_out.mixtures])

    if logging:
        logger.info('')
        logger.info(f'Found {len(mixdb_out.mixtures):,} mixtures to process')
        logger.info(f'{total_samples:,} samples')

    return mixdb_out, total_samples


def genmix(mixdb: MixtureDatabase,
           mixid: MixtureID,
           compute_segsnr: bool = False,
           compute_truth: bool = False,
           logging: bool = False,
           show_progress: bool = False,
           progress: tqdm = None,
           initial_i_sample_offset: int = 0,
           initial_i_frame_offset: int = 0,
           initial_o_frame_offset: int = 0) -> (np.ndarray,
                                                np.ndarray,
                                                np.ndarray,
                                                np.ndarray,
                                                np.ndarray,
                                                MixtureDatabase):
    import ctypes

    from tqdm import tqdm

    import sonusai
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import build_noise_audios
    from sonusai.mixture import build_target_audios
    from sonusai.mixture import get_total_class_count
    from sonusai.mixture import set_mixture_offsets
    from sonusai.utils import human_readable_size
    from sonusai.utils import p_tqdm_map
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import to_numpy_array

    mixdb_out, total_samples = init(mixdb=mixdb, mixid=mixid, logging=logging)

    MP_DICT['mixdb'] = mixdb_out
    MP_DICT['total_samples'] = total_samples
    MP_DICT['compute_truth'] = compute_truth
    MP_DICT['compute_segsnr'] = compute_segsnr
    MP_DICT['target_audios'] = build_target_audios(mixdb=mixdb_out, show_progress=show_progress)
    MP_DICT['noise_audios'] = build_noise_audios(mixdb=mixdb_out, show_progress=show_progress)

    mp_mixture = mp.Array(ctypes.c_int16, total_samples)
    if compute_truth:
        mp_truth_t = mp.Array(ctypes.c_float, total_samples * mixdb_out.num_classes)
    else:
        mp_truth_t = mp.Array(ctypes.c_float, 0)
    mp_target = mp.Array(ctypes.c_int16, total_samples)
    mp_noise = mp.Array(ctypes.c_int16, total_samples)
    if compute_segsnr:
        mp_segsnr = mp.Array(ctypes.c_float, total_samples)
    else:
        mp_segsnr = mp.Array(ctypes.c_float, 0)

    mixture = to_numpy_array(mp_mixture, dtype=np.int16)
    if compute_truth:
        truth_t = to_numpy_array(mp_truth_t, dtype=np.single).reshape((total_samples, mixdb_out.num_classes))
    else:
        truth_t = np.empty(0, dtype=np.single)

    target = to_numpy_array(mp_target, dtype=np.int16)
    noise = to_numpy_array(mp_noise, dtype=np.int16)
    if compute_segsnr:
        segsnr = to_numpy_array(mp_segsnr, dtype=np.single)
    else:
        segsnr = np.empty(0, dtype=np.single)

    # First pass to get offsets
    set_mixture_offsets(mixdb=mixdb_out,
                        initial_i_sample_offset=initial_i_sample_offset,
                        initial_i_frame_offset=initial_i_frame_offset,
                        initial_o_frame_offset=initial_o_frame_offset)

    # Second pass to get mixture and truth_t
    progress_needs_close = False
    if progress is None:
        progress = tqdm(total=len(mixdb_out.mixtures), desc='genmix', disable=not show_progress)
        progress_needs_close = True

    p_tqdm_map(_process_mixture,
               list(range(len(mixdb_out.mixtures))),
               progress=progress,
               initializer=mp_init,
               initargs=(mp_mixture, mp_truth_t, mp_target, mp_noise, mp_segsnr))

    if progress_needs_close:
        progress.close()

    mixdb_out.class_count = get_total_class_count(mixdb_out)

    duration = len(mixture) / sonusai.mixture.SAMPLE_RATE
    if logging:
        logger.info('')
        logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
        logger.info(f'mixture:  {human_readable_size(mixture.nbytes, 1)}')
        if compute_truth:
            logger.info(f'truth_t:  {human_readable_size(truth_t.nbytes, 1)}')
        if compute_segsnr:
            logger.info(f'segsnr:   {human_readable_size(segsnr.nbytes, 1)}')

    return mixture, truth_t, target, noise, segsnr, mixdb_out


def _process_mixture(mixid: int) -> None:
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import get_audio_and_truth_t
    from sonusai.mixture import get_samples_in_mixture
    from sonusai.utils import to_numpy_array

    mixdb: MixtureDatabase = MP_DICT['mixdb']
    total_samples = MP_DICT['total_samples']
    target_audios = MP_DICT['target_audios']
    noise_audios = MP_DICT['noise_audios']
    compute_truth = MP_DICT['compute_truth']
    compute_segsnr = MP_DICT['compute_segsnr']

    mixture = to_numpy_array(mp_mixture, dtype=np.int16)

    truth_t = to_numpy_array(mp_truth_t, dtype=np.single)
    if compute_truth:
        truth_t = truth_t.reshape((total_samples, mixdb.num_classes))

    target = to_numpy_array(mp_target, dtype=np.int16)

    noise = to_numpy_array(mp_noise, dtype=np.int16)

    segsnr = to_numpy_array(mp_segsnr, dtype=np.single)

    record = mixdb.mixtures[mixid]
    first_i_sample_offset = mixdb.mixtures[0].i_sample_offset
    samples = get_samples_in_mixture(mixdb, mixid)

    indices = slice(record.i_sample_offset - first_i_sample_offset,
                    record.i_sample_offset - first_i_sample_offset + samples)

    (mixture[indices],
     truth_t[indices],
     target[indices],
     noise[indices],
     segsnr[indices]) = get_audio_and_truth_t(mixdb=mixdb,
                                              mixture=record,
                                              raw_target_audios=target_audios,
                                              raw_noise_audios=noise_audios,
                                              compute_truth=compute_truth,
                                              compute_segsnr=compute_segsnr)


def main():
    import time
    from os.path import splitext

    import h5py
    from docopt import docopt
    from tqdm import tqdm

    import sonusai
    from sonusai import SonusAIError
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import check_audio_files_exist
    from sonusai.mixture import get_feature_frames_in_mixture
    from sonusai.mixture import get_samples_in_mixture
    from sonusai.mixture import get_transform_frames_in_mixture
    from sonusai.mixture import load_mixdb
    from sonusai.utils import expand_range
    from sonusai.utils import grouper
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixdb_name = args['--mixdb']
    mixid = args['--mixid']
    output_name = args['--output']
    compute_segsnr = args['--segsnr']
    compute_truth = args['--truth']

    if not output_name:
        output_name = splitext(mixdb_name)[0] + '.h5'

    start_time = time.monotonic()

    log_name = 'genmix.log'
    create_file_handler(log_name)
    update_console_handler(verbose)
    initial_log_messages('genmix')

    logger.info(f'\nLoad mixture database from {mixdb_name}')
    mixdb = load_mixdb(name=mixdb_name)
    if mixid == '*':
        mixid = list(range(len(mixdb.mixtures)))
    else:
        mixid = expand_range(mixid)

    mixdb_out, total_samples = init(mixdb=mixdb,
                                    mixid=mixid,
                                    logging=True)

    check_audio_files_exist(mixdb)

    with h5py.File(output_name, 'w') as f:
        f.create_dataset(name='mixture',
                         shape=(total_samples,),
                         dtype=np.int16)
        if compute_truth:
            f.create_dataset(name='truth_t',
                             shape=(total_samples, mixdb_out.num_classes),
                             dtype=np.single)
        f.create_dataset(name='target',
                         shape=(total_samples,),
                         dtype=np.int16)
        f.create_dataset(name='noise',
                         shape=(total_samples,),
                         dtype=np.int16)
        if compute_segsnr:
            f.create_dataset(name='segsnr',
                             shape=(total_samples,),
                             dtype=np.single)

    chunk_size = 100
    progress = tqdm(total=len(mixid), desc='genmix')
    mixid = grouper(range(len(mixdb_out.mixtures)), chunk_size)
    mixdb_out.class_count = [0] * mixdb_out.num_classes

    i_sample_offset = 0
    i_frame_offset = 0
    o_frame_offset = 0
    for m in mixid:
        mixture, truth_t, target, noise, segsnr, mixdb_tmp = genmix(mixdb=mixdb_out,
                                                                    mixid=m,
                                                                    compute_segsnr=compute_segsnr,
                                                                    compute_truth=compute_truth,
                                                                    logging=False,
                                                                    progress=progress,
                                                                    initial_i_sample_offset=i_sample_offset,
                                                                    initial_i_frame_offset=i_frame_offset,
                                                                    initial_o_frame_offset=o_frame_offset)

        samples = mixture.shape[0]
        if compute_truth:
            if samples != truth_t.shape[0]:
                raise SonusAIError(
                    f'truth_t samples does not match mixture samples: {truth_t.shape[0]} != {samples}')
            if mixdb_out.num_classes != truth_t.shape[1]:
                raise SonusAIError(
                    f'truth_t num_classes is incorrect: {truth_t.shape[1]} != {mixdb_out.num_classes}')
        if samples != target.shape[0]:
            raise SonusAIError(f'target samples does not match mixture samples: {target.shape[0]} != {samples}')
        if samples != noise.shape[0]:
            raise SonusAIError(f'noise samples does not match mixture samples: {noise.shape[0]} != {samples}')
        if compute_segsnr and samples != segsnr.shape[0]:
            raise SonusAIError(f'segsnr samples does not match mixture samples: {segsnr.shape[0]} != {samples}')

        with h5py.File(output_name, 'a') as f:
            indices = slice(i_sample_offset, i_sample_offset + samples)

            mixture_dataset = f['mixture']
            mixture_dataset[indices] = mixture

            if compute_truth:
                truth_dataset = f['truth_t']
                truth_dataset[indices] = truth_t

            target_dataset = f['target']
            target_dataset[indices] = target

            noise_dataset = f['noise']
            noise_dataset[indices] = noise

            if compute_segsnr:
                segsnr_dataset = f['segsnr']
                segsnr_dataset[indices] = segsnr

        for idx, val in enumerate(m):
            mixdb_out.mixtures[val] = mixdb_tmp.mixtures[idx]
        for idx in range(mixdb_out.num_classes):
            mixdb_out.class_count[idx] += mixdb_tmp.class_count[idx]

        i_sample_offset = mixdb_out.mixtures[m[-1]].i_sample_offset
        i_sample_offset += get_samples_in_mixture(mixdb_out, m[-1])
        i_frame_offset = mixdb_out.mixtures[m[-1]].i_frame_offset
        i_frame_offset += get_transform_frames_in_mixture(mixdb_out, m[-1])
        o_frame_offset = mixdb_out.mixtures[m[-1]].o_frame_offset
        o_frame_offset += get_feature_frames_in_mixture(mixdb_out, m[-1])

    with h5py.File(output_name, 'a') as f:
        f.attrs['mixdb'] = mixdb_out.to_json()

    progress.close()

    logger.info(f'Wrote {output_name}')
    duration = total_samples / sonusai.mixture.SAMPLE_RATE
    logger.info('')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'mixture:  {human_readable_size(i_sample_offset * 2, 1)}')
    if compute_truth:
        logger.info(f'truth_t:  {human_readable_size(i_sample_offset * mixdb_out.num_classes * 4, 1)}')
    logger.info(f'target:   {human_readable_size(i_sample_offset * 2, 1)}')
    logger.info(f'noise:    {human_readable_size(i_sample_offset * 2, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(i_sample_offset * 4, 1)}')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



"""sonusai keras_predict

usage: keras_predict [-hvfc] (-m MODEL) (-w KMODEL) [-b BATCH] [-t TSTEPS] DATA

options:
   -h, --help
   -v, --verbose                Be verbose.
   -f, --flatten                Flatten input feature data.
   -c, --add1ch                 Add channel dimension to feature (i.e., cnn input).
   -m MODEL, --model MODEL      Model Python file with build and/or hypermodel functions.
   -w KMODEL, --weights KMODEL  Keras model HDF5 weights file, or model file with weights.
   -b BATCH, --batch BATCH      Batch size. [default: 1].
   -t TSTEPS, --tsteps TSTEPS   Timesteps. [default: 0].

Run Keras prediction on a model defined by a Sonusai Keras Python definition file scu_<MODEL>.py
and using DATA HDF5 feature+truth data generated from the SonusAI genft function.

Results are written to HDF5 file with predict and truthf datasets.

Results are written into subdirectory predict-<DATA>-<TIMESTAMP> including two onnx model files:
  - <MODEL>.onnx      onnx model file with batch_size and timesteps equal to provided parameters
  - <MODEL>-b1.onnx   onnx model file with batch_size=1 and if the timesteps dimension exists it
                      is set =1  (b=1,tsteps=1 model is useful for real-time inference applications)

"""
from sonusai import logger


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    import os
    from datetime import datetime

    import h5py
    import keras2onnx
    import keras_tuner as kt
    import numpy as np
    import tensorflow as tf
    from tensorflow.keras import backend as kb

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import load_mixdb
    from sonusai.utils import create_onnx_from_keras
    from sonusai.utils import import_keras_model
    from sonusai.utils import reshape_inputs

    verbose = args['--verbose']
    flatten = args['--flatten']
    add1ch = args['--add1ch']
    model_name = args['--model']
    kmodel_name = args['--weights']
    batch_size = int(args['--batch'])
    timesteps = int(args['--tsteps'])
    data_name = args['DATA']

    model_tail = os.path.basename(model_name)
    data_tail = os.path.basename(data_name)
    data_root = os.path.splitext(data_tail)[0]

    # Results subdirectory
    ts = datetime.now()

    # First try just date
    ts_str = 'predict-' + data_root + '-' + ts.strftime('%Y%m%d')
    try:
        os.mkdir(ts_str)
    except OSError as _:
        # add hour-min-sec if necessary
        ts_str = 'predict-' + data_root + '-' + ts.strftime('%Y%m%d-%H%M%S')
        os.mkdir(ts_str)

    # Setup logging file
    name_ts = ts_str + '/' + 'predict'  # base filename path with subdir timestamp
    create_file_handler(name_ts + '.log')
    update_console_handler(verbose)
    initial_log_messages('predict-' + data_root)

    # Check dims and model build before we read large dataset and make subdir and start logging
    logger.info(f'TF ver: {tf.__version__}')
    logger.info(f'Keras ver: {tf.keras.__version__}')
    logger.info(f'Keras2onnx ver: {keras2onnx.__version__}')

    mixdb = load_mixdb(data_name)
    mixid = list(range(len(mixdb.mixtures)))
    logger.info(f'Found {len(mixid)} mixtures with {mixdb.num_classes} classes from {data_name}')

    # Import model definition file
    model = import_keras_model(model_name)
    logger.info(f'Successfully imported {model_tail}, testing default model build')

    # Calculate input shape
    logger.info('Building default model')

    try:
        hypermodel = model.MyHyperModel(num_classes=mixdb.num_classes,
                                        feature=mixdb.feature,
                                        timesteps=timesteps,
                                        flatten=flatten,
                                        add1ch=add1ch,
                                        batch_size=batch_size)
        model_default = hypermodel.build_model(kt.HyperParameters())
    except Exception as e:
        logger.exception(f'Error: build_model() in {model_tail} failed: {e}.')
        raise SystemExit(1)

    logger.info(f'Successfully built using {model_tail}, summary:')
    kb.clear_session()
    logger.info('')
    model_default.summary(print_fn=logger.info)
    logger.info(f'User shape parameters: batch_size {batch_size}, timesteps {timesteps}, '
                f'flatten={flatten}, add1ch={add1ch}')
    logger.info(f'Model build above with default hyper-parameters, in_shape: {hypermodel.input_shape}, '
                f'num_classes {mixdb.num_classes}')
    logger.info(f'Compiled with optimizer: {model_default.optimizer.get_config()}')
    logger.info('')

    logger.info(f'Loading weights from {kmodel_name}')
    model_default.load_weights(kmodel_name)
    logger.info(f'Loading feature+truth data from {data_name} with {len(mixid)} mixtures '
                f'and {mixdb.num_classes} classes')
    with h5py.File(data_name, 'r') as f:
        feature = np.array(f['feature'])
        truth_f = np.array(f['truth_f'])

    logger.info('Reshaping feature and truth data')
    feature, truth, in_shape, outlen, _, _ = reshape_inputs(feature, truth_f, batch_size, timesteps, flatten, add1ch)

    # Save to ONNX format with specified batch and timestep sizes
    logger.info('Saving model to onnx file {}'.format(name_ts + '.onnx'))
    try:
        create_onnx_from_keras(keras_model=model_default,
                               is_flattened=flatten,
                               has_timestep=(timesteps != 0),
                               has_channel=add1ch,
                               is_mutex=mixdb.truth_mutex,
                               feature=mixdb.feature,
                               filename=name_ts + '.onnx')
    except Exception as e:
        logger.info(f'Failed to create ONNX model, no file saved: {e}.')

    # Compute prediction and metrics on data
    logger.info('Running Keras prediction')
    predict = model_default.predict(feature, batch_size=batch_size, verbose=1)
    with h5py.File(name_ts + '.h5', 'w') as f:
        f.create_dataset('predict', data=predict)
    logger.info('Saved prediction data to {}'.format(name_ts + '.h5'))

    # Create and save model with timesteps, batch = 1
    #logger.info('')
    if timesteps > 0:
        # only set to 1 if nonzero (exists)
        timesteps = 1

    hypermodel = model.MyHyperModel(num_classes=mixdb.num_classes,
                                    feature=mixdb.feature,
                                    timesteps=timesteps,
                                    flatten=flatten,
                                    add1ch=add1ch,
                                    batch_size=1)

    modelp = hypermodel.build_model(kt.HyperParameters())
    # load weights from previously saved HDF5
    modelp.load_weights(kmodel_name)

    # save a prediction version of model to name_ts-predict-onnx
    create_onnx_from_keras(keras_model=modelp,
                           is_flattened=flatten,
                           has_timestep=(timesteps != 0),
                           has_channel=add1ch,
                           is_mutex=mixdb.truth_mutex,
                           feature=mixdb.feature,
                           filename=name_ts + '-b1.onnx')
    logger.info('Saved inference model (batch_size=1) to onnx file {}'.format(name_ts + '-b1.onnx'))

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



from dataclasses_json import DataClassJsonMixin


class DataClassSonusAIMixin(DataClassJsonMixin):
    from typing import Dict
    from typing import Union

    # Json type defined to maintain compatibility with DataClassJsonMixin
    Json = Union[dict, list, str, int, float, bool, None]

    def __str__(self):
        return f'{self.to_dict()}'

    # Override DataClassJsonMixin to remove dictionary keys with values of None
    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        def del_none(d):
            if isinstance(d, dict):
                for key, value in list(d.items()):
                    if value is None:
                        del d[key]
                    elif isinstance(value, dict):
                        del_none(value)
                    elif isinstance(value, list):
                        for item in value:
                            del_none(item)
            elif isinstance(d, list):
                for item in d:
                    del_none(item)
            return d

        return del_none(super().to_dict(encode_json))



from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np

from sonusai.mixture.dataclasses_sonusai import DataClassSonusAIMixin
from sonusai.mixture.segment import Segment


@dataclass(frozen=True)
class TruthSetting(DataClassSonusAIMixin):
    index: Optional[List[int]] = None
    function: Optional[str] = None
    config: Optional[dict] = None


TruthSettings = List[TruthSetting]

OptionalNumberStr = Optional[Union[float, int, str]]
OptionalListNumberStr = Optional[List[Union[float, int, str]]]


@dataclass
class Augmentation(DataClassSonusAIMixin):
    normalize: OptionalNumberStr = None
    pitch: OptionalNumberStr = None
    tempo: OptionalNumberStr = None
    gain: OptionalNumberStr = None
    eq1: OptionalListNumberStr = None
    eq2: OptionalListNumberStr = None
    eq3: OptionalListNumberStr = None
    lpf: OptionalNumberStr = None
    count: Optional[int] = None
    mixup: Optional[int] = 1


Augmentations = List[Augmentation]


@dataclass(frozen=True)
class TargetFile(DataClassSonusAIMixin):
    name: str
    duration: float
    truth_settings: TruthSettings
    augmentations: Optional[Augmentations] = None
    class_balancing_augmentation: Optional[Augmentation] = None


TargetFiles = List[TargetFile]


@dataclass
class AugmentedTarget(DataClassSonusAIMixin):
    target_file_index: int
    target_augmentation_index: int


AugmentedTargets = List[AugmentedTarget]


@dataclass(frozen=True)
class NoiseFile(DataClassSonusAIMixin):
    name: str
    duration: float
    augmentations: Optional[Augmentations] = None


NoiseFiles = List[NoiseFile]

ClassCount = List[int]


@dataclass
class Mixture(DataClassSonusAIMixin):
    target_file_index: List[int] = None
    noise_file_index: int = None
    noise_offset: int = None
    target_augmentation_index: List[int] = None
    noise_augmentation_index: int = None
    snr: float = None
    random_snr: bool = None
    samples: int = None
    target_gain: List[int] = None
    class_count: ClassCount = None
    target_snr_gain: float = None
    noise_snr_gain: float = None
    i_sample_offset: Optional[int] = None
    i_frame_offset: Optional[int] = None
    o_frame_offset: Optional[int] = None


Mixtures = List[Mixture]

MixtureID = Union[str, List[int]]


@dataclass
class MixtureDatabase(DataClassSonusAIMixin):
    class_balancing: Optional[bool] = False
    class_balancing_augmentation: Optional[Augmentation] = None
    class_count: ClassCount = None
    class_labels: List[str] = None
    class_weights_threshold: List[float] = None
    exhaustive_noise: Optional[bool] = True
    feature: str = None
    feature_samples: int = None
    feature_step_samples: int = None
    first_cba_index: Optional[int] = None
    frame_size: int = None
    mixtures: Mixtures = None
    noise_augmentations: Augmentations = None
    noises: NoiseFiles = None
    num_classes: int = None
    random_snrs: List[str] = None
    seed: Optional[int] = 0
    snrs: List[float] = None
    target_augmentations: Augmentations = None
    targets: TargetFiles = None
    truth_mutex: bool = None
    truth_reduction_function: str = None
    truth_settings: TruthSettings = None


def mixdb_json_from_file(name: str) -> str:
    from os.path import exists
    from os.path import splitext

    import h5py

    from sonusai import SonusAIError

    if not exists(name):
        raise SonusAIError(f'{name} does not exist')

    ext = splitext(name)[1]

    if ext == '.json':
        with open(file=name, mode='r', encoding='utf-8') as f:
            return f.read()

    if ext == '.h5':
        with h5py.File(name=name, mode='r') as f:
            return f.attrs['mixdb']

    raise SonusAIError(f'Do not know how to load mixdb from {name}')


def mixdb_from_json(data: str) -> MixtureDatabase:
    import json

    return MixtureDatabase.from_dict(json.loads(data))


def load_mixdb(name: str) -> MixtureDatabase:
    return mixdb_from_json(mixdb_json_from_file(name))


def set_mixture_offsets(mixdb: MixtureDatabase,
                        initial_i_sample_offset: int = 0,
                        initial_i_frame_offset: int = 0,
                        initial_o_frame_offset: int = 0) -> None:
    i_sample_offset = initial_i_sample_offset
    i_frame_offset = initial_i_frame_offset
    o_frame_offset = initial_o_frame_offset
    for mixid in range(len(mixdb.mixtures)):
        mixdb.mixtures[mixid].i_sample_offset = i_sample_offset
        mixdb.mixtures[mixid].i_frame_offset = i_frame_offset
        mixdb.mixtures[mixid].o_frame_offset = o_frame_offset

        i_sample_offset += get_samples_in_mixture(mixdb, mixid)
        i_frame_offset += get_transform_frames_in_mixture(mixdb, mixid)
        o_frame_offset += get_feature_frames_in_mixture(mixdb, mixid)


def get_samples_in_mixture(mixdb: MixtureDatabase, mixid: int) -> int:
    return mixdb.mixtures[mixid].samples


def get_transform_frames_in_mixture(mixdb: MixtureDatabase, mixid: int) -> int:
    return mixdb.mixtures[mixid].samples // mixdb.frame_size


def get_feature_frames_in_mixture(mixdb: MixtureDatabase, mixid: int) -> int:
    return mixdb.mixtures[mixid].samples // mixdb.feature_step_samples


def get_sample_offsets_in_mixture(mixdb: MixtureDatabase, mixid: int) -> (int, int):
    i_sample_offset = sum([sub.samples for sub in mixdb.mixtures[:mixid]])
    return i_sample_offset, i_sample_offset + mixdb.mixtures[mixid].samples


def get_transform_frame_offsets_in_mixture(mixdb: MixtureDatabase, mixid: int) -> (int, int):
    start, stop = get_sample_offsets_in_mixture(mixdb, mixid)
    return start // mixdb.frame_size, stop // mixdb.frame_size


def get_feature_frame_offsets_in_mixture(mixdb: MixtureDatabase, mixid: int) -> (int, int):
    start, stop = get_sample_offsets_in_mixture(mixdb, mixid)
    return start // mixdb.feature_step_samples, stop // mixdb.feature_step_samples


def get_class_weights_threshold(mixdb: Union[MixtureDatabase, Dict]) -> List[float]:
    """Get the class_weights_threshold from a mixture database or a config."""

    from sonusai import SonusAIError

    if isinstance(mixdb, dict):
        class_weights_threshold = mixdb['class_weights_threshold']
        num_classes = mixdb['num_classes']
    else:
        class_weights_threshold = mixdb.class_weights_threshold
        num_classes = mixdb.num_classes

    if not isinstance(class_weights_threshold, list):
        class_weights_threshold = [class_weights_threshold] * num_classes

    if len(class_weights_threshold) != num_classes:
        raise SonusAIError(f'invalid class_weights_threshold length: {len(class_weights_threshold)}')

    return class_weights_threshold


def get_file_frame_segments(mixdb: MixtureDatabase, mixid: MixtureID = ':') -> Dict[int, Segment]:
    """
    Get Segment data from a mixture database
    Return a dictionary where:
        - keys are the mixids
        - values are file frame Segment objects
    """
    _mixid = convert_mixid_to_list(mixdb, mixid)

    # remove duplicates
    _mixid = sorted(list(set(_mixid)))

    file_frame_segments = dict()
    for m in _mixid:
        file_frame_segments[m] = Segment(mixdb.mixtures[m].o_frame_offset,
                                         get_feature_frames_in_mixture(mixdb, m))

    return file_frame_segments


def get_offsets(mixdb: MixtureDatabase, mixid: int) -> (int, int, int):
    from sonusai import SonusAIError

    if mixid >= len(mixdb.mixtures) or mixid < 0:
        raise SonusAIError(f'Invalid mixid: {mixid}')

    i_sample_offset = sum([sub.samples for sub in mixdb.mixtures[:mixid]])
    i_frame_offset = i_sample_offset // mixdb.frame_size
    o_frame_offset = i_sample_offset // mixdb.feature_step_samples

    return i_sample_offset, i_frame_offset, o_frame_offset


def new_mixdb_from_mixid(mixdb: MixtureDatabase, mixid: MixtureID) -> MixtureDatabase:
    from copy import deepcopy

    from sonusai import SonusAIError

    mixdb_out = deepcopy(mixdb)
    mixdb_out.mixtures = get_mixtures_from_mixid(mixdb_out, mixid)
    set_mixture_offsets(mixdb_out)

    if not mixdb_out.mixtures:
        raise SonusAIError(f'Error processing mixid: {mixid}; resulted in empty list of mixtures')

    return mixdb_out


def load_mixid(mixdb: MixtureDatabase, name: str = None) -> List[int]:
    import json
    from os.path import exists

    from sonusai import SonusAIError

    if name is None:
        mixid = list(range(len(mixdb.mixtures)))
    else:
        if not exists(name):
            raise SonusAIError(f'{name} does not exist')

        with open(file=name, mode='r', encoding='utf-8') as f:
            mixid = json.load(f)
            if not isinstance(mixid, dict) or 'mixid' not in mixid:
                raise SonusAIError(f'Could not find ''mixid'' in {name}')
            mixid = mixid['mixid']

    return mixid


def convert_mixid_to_list(mixdb: MixtureDatabase, mixid: MixtureID = None) -> List[int]:
    mixid_out = mixid

    if mixid_out is None:
        return list(range(len(mixdb.mixtures)))

    if isinstance(mixid_out, str):
        try:
            mixid_out = eval(f'{list(range(len(mixdb.mixtures)))}[{mixid_out}]')
        except NameError:
            return []

    if not all(isinstance(x, int) and x < len(mixdb.mixtures) for x in mixid_out):
        return []

    return mixid_out


def get_mixtures_from_mixid(mixdb: MixtureDatabase, mixid: MixtureID = None) -> Mixtures:
    from copy import deepcopy

    return [deepcopy(mixdb.mixtures[i]) for i in convert_mixid_to_list(mixdb, mixid)]


def get_mixid_data(mixdb: MixtureDatabase,
                   mixid: MixtureID,
                   truth_f: np.ndarray,
                   predict: np.ndarray) -> (np.ndarray, np.ndarray):
    """Collect per-feature data of specified mixids from mixdb where inputs are:
       truth_f:   truth data matching mixdb (size feature_frames x num_classes)
       predict:   prediction or segsnr data (size feature_frames x ndim, where ndim > 1)

    Returns:
        truth_f_out:    np.ndarray combined truth from mixids
        predict_out:    np.ndarray combined data from mixids
    """
    from sonusai import SonusAIError
    from sonusai.mixture import get_slices

    in_frames = truth_f.shape[0]

    if in_frames != predict.shape[0]:
        raise SonusAIError('truth_f and predict have different number of frames')

    num_classes = truth_f.shape[1]
    # same as num_class for prediction data, but use for segsnr too
    num_data = predict.shape[1]

    _mixid = convert_mixid_to_list(mixdb, mixid)
    file_frame_segments = get_file_frame_segments(mixdb, _mixid)
    out_frames = sum([file_frame_segments[m].length for m in _mixid])
    truth_f_out = np.empty((out_frames, num_classes), dtype=np.single)
    predict_out = np.empty((out_frames, num_data), dtype=np.single)

    # Handle the special case when input data is smaller, i.e., prediction when total mixture
    # length is a non-multiple of the batch size. In this case just pad both with zeros; should
    # have negligible effect on metrics
    max_in_index = max([file_frame_segments[m].get_slice().stop for m in _mixid])
    frame_diff = max_in_index - in_frames
    if frame_diff > 0:
        truth_f = np.concatenate((truth_f, np.zeros((frame_diff, num_classes))))
        predict = np.concatenate((predict, np.zeros((frame_diff, num_classes))))

    start = 0
    for m in _mixid:
        in_slice, out_slice = get_slices(file_frame_segments[m], start)

        if len(truth_f[in_slice]) != len(truth_f_out[out_slice]):
            raise SonusAIError(f'in_slice and out_slice are different lengths')

        truth_f_out[out_slice] = truth_f[in_slice]
        predict_out[out_slice] = predict[in_slice]
        start += file_frame_segments[m].length

    return truth_f_out, predict_out


def evaluate_random_rule(rule: str) -> Union[str, float]:
    """Evaluate 'rand' directive."""
    import re
    from random import uniform

    from sonusai.mixture import RAND_PATTERN

    def rand_repl(m):
        return f'{uniform(float(m.group(1)), float(m.group(4))):.2f}'

    return eval(re.sub(RAND_PATTERN, rand_repl, rule))



def grouper(iterable, n):
    from itertools import zip_longest

    args = [iter(iterable)] * n
    result = zip_longest(*args, fillvalue=None)
    return [list(filter(None.__ne__, x)) for x in result]



from typing import List
from typing import Union

import numpy as np
import pandas as pd

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureID


def confusion_matrix_summary(mixdb: MixtureDatabase,
                             mixid: MixtureID,
                             truth_f: np.ndarray,
                             predict: np.ndarray,
                             class_idx: int,
                             pred_thr: Union[float, List[float]] = 0,
                             truth_thr: float = 0.5,
                             timesteps: int = 0) -> (pd.DataFrame, pd.DataFrame):
    """Calculate confusion matrix for specified class, using truth and prediction
       data features x num_classes.

       pred_thr sets the decision threshold(s) applied to predict data, thus allowing
       predict to be continuous probabilities.

       Default pred_thr=0 will infer 0.5 for multi-label mode (truth_mutex = False), or
       if single-label mode (truth_mutex == True) then ignore and use argmax mode, and
       the confusion matrix is calculated for all classes.

       Returns pandas dataframes of confusion matrix cmdf and normalized confusion matrix cmndf.
    """
    from sonusai.metrics import one_hot
    from sonusai.mixture import get_mixid_data

    num_classes = truth_f.shape[1]
    ytrue, ypred = get_mixid_data(mixdb=mixdb, mixid=mixid, truth_f=truth_f, predict=predict)

    # Check pred_thr array or scalar and return final scalar pred_thr value
    if not mixdb.truth_mutex and num_classes > 1:
        if np.ndim(pred_thr) == 0 and pred_thr == 0:
            # multi-label pred_thr scalar 0 force to 0.5 default
            pred_thr = 0.5

        if np.ndim(pred_thr) == 1:
            if len(pred_thr) == 1:
                if pred_thr[0] == 0:
                    # multi-label pred_thr array scalar 0 force to 0.5 default
                    pred_thr = 0.5
                else:
                    # multi-label pred_thr array set to scalar = array[0]
                    pred_thr = pred_thr[0]
            else:
                # multi-label pred_thr array scalar set = array[class_idx]
                pred_thr = pred_thr[class_idx]

    if len(mixdb.class_labels) == num_classes:
        class_names = mixdb.class_labels
    else:
        class_names = ([f'Class {i}' for i in range(1, num_classes + 1)])

    class_nums = ([f'{i}' for i in range(1, num_classes + 1)])

    if mixdb.truth_mutex:
        # single-label mode force to argmax mode
        pred_thr = 0
        _, _, cm, cmn, _, _ = one_hot(ytrue, ypred, pred_thr, truth_thr, timesteps)
        row_n = class_names
        row_n[-1] = 'Other'
        # mux = pd.MultiIndex.from_product([['Single-label/mutex mode, truth thr = {}'.format(truth_thr)],
        #                                   class_nums])
        # mux = pd.MultiIndex.from_product([['truth thr = {}'.format(truth_thr)], class_nums])

        cmdf = pd.DataFrame(cm, index=row_n, columns=class_nums, dtype=int)
        cmndf = pd.DataFrame(cmn, index=row_n, columns=class_nums, dtype=float)

    else:
        _, _, cm, cmn, _, _ = one_hot(ytrue[:, class_idx], ypred[:, class_idx], pred_thr, truth_thr, timesteps)
        cname = class_names[class_idx]
        row_n = ['TrueN', 'TrueP']
        col_n = ['N-' + cname, 'P-' + cname]
        cmdf = pd.DataFrame(cm, index=row_n, columns=col_n, dtype=int)
        cmndf = pd.DataFrame(cmn, index=row_n, columns=col_n, dtype=float)
        # add thresholds in 3rd row
        pdnote = pd.DataFrame(np.atleast_2d([pred_thr, truth_thr]), index=['p/t thr:'], columns=col_n)
        cmdf = pd.concat([cmdf, pdnote])
        cmndf = pd.concat([cmndf, pdnote])

    return cmdf, cmndf



"""sonusai evaluate

usage: evaluate [-hv] (-f FEATURE) (-n PREDICT) [-t PTHR]

options:
   -h, --help
   -v, --verbose                    Be verbose.
   -f FEATURE, --feature FEATURE    Feature + truth .h5 data file
   -n PREDICT, --predict PREDICT    Predict .h5 data file.
   -t PTHR, --thr PTHR              Optional prediction decision threshold(s). [default: 0].

Evaluate calculates performance metrics of neural-network models from model prediction data and the associated
feature+truth data file.

Inputs:
    FEATURE     A SonusAI feature+truth HDF5 file. Contains:
                    attribute:  mixdb
                    dataset:    feature   Note: not required for metrics
                    dataset:    truth_f
                    dataset:    segsnr    Optional
    PREDICT     A SonusAI predict HDF5 file. Contains:
                    dataset:    predict (either [frames, num_classes] or [frames, timesteps, num_classes])
    PTHR        Scalar or array of thresholds. Default 0 will select values:
                    argmax()    if mixdb indicates single-label mode (truth_mutex = true)
                    0.5         if mixdb indicates multi-label mode  (truth_mutex = false)
                If PTHR = -1, optimal thresholds are calculated using precision_recall_curve() which
                optimizes F1 score.
"""
from pathlib import Path
from typing import Union

import numpy as np

from sonusai import logger
from sonusai.mixture import MixtureDatabase


def evaluate(mixdb: MixtureDatabase,
             truth: np.ndarray,
             predict: Union[None, np.ndarray],
             segsnr: Union[None, np.ndarray] = None,
             output_dir: Union[None, str, Path] = None,
             predict_thr: Union[float, np.ndarray] = 0.0,
             feature: Union[None, np.ndarray] = None,
             verbose: bool = False) -> None:
    from os.path import join

    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.metrics import calculate_optimal_thresholds
    from sonusai.metrics import class_summary
    from sonusai.metrics import snr_summary
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.queries import get_mixids_from_snr
    from sonusai.utils import human_readable_size
    from sonusai.utils import reshape_truth_predict
    from sonusai.utils import seconds_to_hms

    update_console_handler(verbose)
    initial_log_messages('evaluate')

    if truth.shape[-1] != predict.shape[-1]:
        logger.exception(f'Number of classes in truth and predict are not equal. Exiting.')
        raise SystemExit(1)

    # truth, predict can be either frames x num_classes, or frames x timesteps x num_classes
    # in binary case dim may not exist, detect this and set num_classes == 1
    timesteps = -1
    truth, predict, num_classes = reshape_truth_predict(truth, predict, timesteps)

    fdiff = truth.shape[0] - predict.shape[0]
    if fdiff > 0:
        # truth = truth[0:-fdiff,:]
        predict = np.concatenate((predict, np.zeros((fdiff, num_classes))))
        logger.info(f'Truth has more feature-frames than predict, padding predict with zeros to match.')

    if fdiff < 0:
        predict = predict[0:fdiff, :]
        logger.info(f'Predict has more feature-frames than truth, trimming predict to match.')

    # Check segsnr, input is always in transform frames
    compute_segsnr = False
    if len(segsnr) > 0:
        segsnr_feature_frames = segsnr.shape[0] / (mixdb.feature_step_samples / mixdb.frame_size)
        if segsnr_feature_frames == truth.shape[0]:
            compute_segsnr = True
        else:
            logger.warning('segsnr length does not match truth, ignoring.')

    # Check pred_thr array or scalar and return final scalar pred_thr value
    if not mixdb.truth_mutex:
        if num_classes > 1:
            if np.ndim(predict_thr) == 0 and predict_thr == 0:
                # multi-label pred_thr scalar 0 force to 0.5 default
                predict_thr = 0.5

            if np.ndim(predict_thr) == 1:
                if len(predict_thr) == 1:
                    if predict_thr[0] == 0:
                        # multi-label pred_thr array scalar 0 force to 0.5 default
                        predict_thr = 0.5
                    else:
                        # multi-label pred_thr array set to scalar = array[0]
                        predict_thr = predict_thr[0]

    else:
        # single-label mode, force argmax mode
        predict_thr = 0

    if predict_thr == -1:
        thrpr, thrroc, _, _ = calculate_optimal_thresholds(truth, predict, timesteps)
        predict_thr = np.atleast_1d(thrpr)
        predict_thr = np.maximum(predict_thr, 0.001)  # enforce lower limit
        predict_thr = np.minimum(predict_thr, 0.999)  # enforce upper limit
        predict_thr = predict_thr.round(2)

    # Summarize the mixture data
    num_mixtures = len(mixdb.mixtures)
    total_samples = sum([sub.samples for sub in mixdb.mixtures])
    duration = total_samples / SAMPLE_RATE

    logger.info('')
    logger.info(f'Mixtures: {num_mixtures}')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'truth:    {human_readable_size(truth.nbytes, 1)}')
    logger.info(f'predict:  {human_readable_size(predict.nbytes, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(segsnr.nbytes, 1)}')
    if feature:
        logger.info(f'feature:  {human_readable_size(feature.nbytes, 1)}')

    logger.info(f'Classes: {num_classes}')
    if mixdb.truth_mutex:
        logger.info(f'Mode:  Single-label / truth_mutex / softmax')
    else:
        logger.info(f'Mode:  Multi-label / Binary')

    mxid_snro = get_mixids_from_snr(mixdb=mixdb)
    snrlist = list(mxid_snro.keys())
    snrlist.sort(reverse=True)
    logger.info(f'Ordered SNRs: {snrlist}\n')
    if type(predict_thr) is np.ndarray:
        logger.info(f'Prediction Threshold(s): {predict_thr.transpose()}\n')
    else:
        logger.info(f'Prediction Threshold(s): {predict_thr}\n')

    # Top-level report over all mixtures
    macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb=mixdb,
                                                      mixid=':',
                                                      truth_f=truth,
                                                      predict=predict,
                                                      segsnr=segsnr if compute_segsnr else None,
                                                      predict_thr=predict_thr)

    if num_classes > 1:
        logger.info(f'Metrics micro-avg per SNR over all {num_mixtures} mixtures:')
    else:
        logger.info(f'Metrics per SNR over all {num_mixtures} mixtures:')
    logger.info(microdf.round(3).to_string())
    logger.info('')
    if output_dir:
        microdf.round(3).to_csv(join(output_dir, 'snr.csv'))

    if mixdb.truth_mutex:
        macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb=mixdb,
                                                          mixid=':',
                                                          truth_f=truth[:, 0:-1],
                                                          predict=predict[:, 0:-1],
                                                          segsnr=segsnr if compute_segsnr else None,
                                                          predict_thr=predict_thr)

        logger.info(f'Metrics micro-avg without "Other" class per SNR over all {num_mixtures} mixtures:')
        logger.info(microdf.round(3).to_string())
        logger.info('')
        if output_dir:
            microdf.round(3).to_csv(join(output_dir, 'snrwo.csv'))

    for snri in snrlist:
        mxids = mxid_snro[snri]
        classdf = class_summary(mixdb, mxids, truth, predict, predict_thr)
        logger.info(f'Metrics per class for SNR {snri} over {len(mxids)} mixtures:')
        logger.info(classdf.round(3).to_string())
        logger.info('')
        if output_dir:
            classdf.round(3).to_csv(join(output_dir, f'class_snr{snri}.csv'))


def main():
    from datetime import datetime
    from os import mkdir

    import h5py
    from docopt import docopt

    import sonusai
    from sonusai import create_file_handler
    from sonusai.mixture import mixdb_from_json
    from sonusai.utils import read_predict_data
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    feature_name = args['--feature']
    predict_name = args['--predict']
    predict_threshold = float(args['--thr'])

    # create output directory
    output_dir = f'evaluate-{datetime.now():%Y%m%d}'
    try:
        mkdir(output_dir)
    except OSError as _:
        output_dir = f'evaluate-{datetime.now():%Y%m%d-%H%M%S}'
        try:
            mkdir(output_dir)
        except OSError as error:
            logger.error(f'Could not create directory, {output_dir}: {error}')
            raise SystemExit(1)

    log_name = output_dir + '/evaluate.log'
    create_file_handler(log_name)

    # mixdb, feature, truth, segsnr = read_feature_data(feature_name)
    with h5py.File(feature_name, 'r') as f:
        mixdb = mixdb_from_json(f.attrs['mixdb'])
        # feature = f['/feature'][:]  # No feature load saves significant time
        truth_f = np.array(f['/truth_f'])
        segsnr = np.array(f['/segsnr'])

    # mixture, target, noise = read_mixture_data(mixture_name)
    predict = read_predict_data(predict_name)

    evaluate(mixdb=mixdb,
             truth=truth_f,
             segsnr=segsnr,
             output_dir=output_dir,
             predict=predict,
             predict_thr=predict_threshold,
             verbose=args['--verbose'])

    logger.info(f'Wrote results to {output_dir}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



def create_onnx_from_keras(keras_model,
                           is_flattened: bool = True,
                           has_timestep: bool = True,
                           has_channel: bool = False,
                           is_mutex: bool = True,
                           feature: str = '',
                           filename: str = ''):
    import tf2onnx

    onnx_model, _ = tf2onnx.convert.from_keras(keras_model)
    onnx_model = replace_stateful_grus(keras_model=keras_model, onnx_model=onnx_model)
    onnx_model = add_sonusai_metadata(model=onnx_model,
                                      is_flattened=is_flattened,
                                      has_timestep=has_timestep,
                                      has_channel=has_channel,
                                      is_mutex=is_mutex,
                                      feature=feature)
    if filename:
        import onnx
        onnx.save(onnx_model, filename)
    return onnx_model


def replace_stateful_grus(keras_model, onnx_model):
    """Replace stateful GRUs with custom layers."""
    from tensorflow.keras.layers import GRU

    stateful_gru_names = []
    for i in range(len(keras_model.layers)):
        layer = keras_model.layers[i]
        if isinstance(layer, GRU):
            if layer.stateful:
                stateful_gru_names.append(layer.name)

    for node_index in range(len(onnx_model.graph.node)):
        node = onnx_model.graph.node[node_index]
        replace = False
        if node.op_type == 'GRU':
            for i in node.input:
                for n in stateful_gru_names:
                    if n in i:
                        replace = True
        if node.name in stateful_gru_names or replace:
            node.op_type = 'SGRU'

    return onnx_model


def add_sonusai_metadata(model,
                         is_flattened: bool = True,
                         has_timestep: bool = True,
                         has_channel: bool = False,
                         is_mutex: bool = True,
                         feature: str = ''):
    """Add SonusAI metadata to ONNX model.
      model           keras model
      file_pfx        filename prefix to save onnx model (do not save if empty)
      is_flattened    model feature data is flattened
      has_timestep    model has timestep dimension
      has_channel     model has channel dimension
      is_mutex        model label output is mutually exclusive
      feature         model feature type
    """
    f_flag = model.metadata_props.add()
    f_flag.key = 'is_flattened'
    f_flag.value = str(is_flattened)

    t_flag = model.metadata_props.add()
    t_flag.key = 'has_timestep'
    t_flag.value = str(has_timestep)

    c_flag = model.metadata_props.add()
    c_flag.key = 'has_channel'
    c_flag.value = str(has_channel)

    m_flag = model.metadata_props.add()
    m_flag.key = 'is_mutex'
    m_flag.value = str(is_mutex)

    feature_flag = model.metadata_props.add()
    feature_flag.key = 'feature'
    feature_flag.value = str(feature)

    return model



def seconds_to_hms(seconds: float) -> str:
    """Convert given seconds into string of H:MM:SS"""
    h = int(seconds / 3600)
    s = seconds - h * 3600
    m = int(s / 60)
    s = int(seconds - h * 3600 - m * 60)
    return f'{h:d}:{m:02d}:{s:02d} (H:MM:SS)'



from typing import List

import numpy as np

from sonusai.mixture.mixdb import Mixture
from sonusai.mixture.mixdb import MixtureDatabase


def _process_target_audio(mixdb: MixtureDatabase,
                          mixture: Mixture,
                          raw_target_audios: List[np.ndarray]) -> List[np.ndarray]:
    """Apply augmentation and update target metadata."""
    from sonusai import SonusAIError
    from sonusai.mixture import apply_augmentation
    from sonusai.mixture import pad_to_samples

    target_file_index = mixture.target_file_index
    target_augmentation_index = mixture.target_augmentation_index
    if len(target_file_index) != len(target_augmentation_index):
        raise SonusAIError('target_file_index and target_augmentation_index are not the same length')

    mixture.samples = 0
    target_audios = list()
    mixture.target_gain = list()
    for idx in range(len(target_file_index)):
        target_augmentation = mixdb.target_augmentations[target_augmentation_index[idx]]

        target_audios.append(apply_augmentation(audio_in=raw_target_audios[target_file_index[idx]],
                                                augmentation=target_augmentation,
                                                length_common_denominator=mixdb.feature_step_samples))

        # target_gain is used to back out the gain augmentation in order to return the target audio
        # to its normalized level when calculating truth.
        if target_augmentation.gain is not None:
            target_gain = 10 ** (target_augmentation.gain / 20)
        else:
            target_gain = 1
        mixture.target_gain.append(target_gain)

    mixture.samples = max([len(item) for item in target_audios])

    for idx in range(len(target_audios)):
        target_audios[idx] = pad_to_samples(audio_in=target_audios[idx], samples=mixture.samples)

    return target_audios


def process_target(mixture: Mixture,
                   mixdb: MixtureDatabase,
                   raw_target_audios: List[np.ndarray]) -> Mixture:
    """Apply augmentation and update target metadata."""
    _process_target_audio(mixdb, mixture, raw_target_audios)
    return mixture


def process_mixture(mixture: Mixture,
                    mixdb: MixtureDatabase,
                    raw_target_audios: List[np.ndarray],
                    augmented_noise_audio: np.ndarray) -> Mixture:
    from sonusai.mixture import get_class_count
    from sonusai.mixture import get_next_noise

    augmented_target_audios = _process_target_audio(mixdb=mixdb,
                                                    mixture=mixture,
                                                    raw_target_audios=raw_target_audios)

    noise_segment, _ = get_next_noise(offset_in=mixture.noise_offset,
                                      length=mixture.samples,
                                      audio_in=augmented_noise_audio)

    mixture = _update_mixture_gains(mixdb=mixdb,
                                    mixture=mixture,
                                    target_audios=augmented_target_audios,
                                    noise_audio=noise_segment)

    for idx in range(len(augmented_target_audios)):
        augmented_target_audios[idx] = np.int16(
            np.round(np.single(augmented_target_audios[idx]) * mixture.target_snr_gain))

    mixture.class_count = get_class_count(mixdb=mixdb,
                                          mixture=mixture,
                                          target_audios=augmented_target_audios,
                                          noise_audio=noise_segment)

    return mixture


def process_mixture_nen(mixture: Mixture,
                        mixdb: MixtureDatabase,
                        raw_target_audios: List[np.ndarray],
                        augmented_noise_audios: List[List[np.ndarray]]) -> Mixture:
    augmented_noise_audio = augmented_noise_audios[mixture.noise_file_index][mixture.noise_augmentation_index]
    return process_mixture(mixture=mixture,
                           mixdb=mixdb,
                           raw_target_audios=raw_target_audios,
                           augmented_noise_audio=augmented_noise_audio)


def _update_mixture_gains(mixdb: MixtureDatabase,
                          mixture: Mixture,
                          target_audios: List[np.ndarray],
                          noise_audio: np.ndarray) -> Mixture:
    from sonusai.utils import dbfs_to_sample

    target_audio = np.single(sum(target_audios))

    if mixture.snr < -96:
        # Special case for zeroing out target data
        mixture.target_snr_gain = 0
        mixture.noise_snr_gain = 1
        mixture.class_count = [0] * mixdb.num_classes
        # Setting target_gain to zero will cause the truth to be all zeros.
        mixture.target_gain = [0] * len(mixture.target_gain)
    elif mixture.snr > 96:
        # Special case for zeroing out noise data
        mixture.target_snr_gain = 1
        mixture.noise_snr_gain = 0
    else:
        target_energy = np.mean(np.square(target_audio))
        noise_energy = np.mean(np.square(np.single(noise_audio)))
        if noise_energy == 0:
            noise_gain = 1
        else:
            noise_gain = np.sqrt(target_energy / noise_energy) / 10 ** (mixture.snr / 20)

        # Check for noise_gain > 1 to avoid clipping
        if noise_gain > 1:
            mixture.target_snr_gain = 1 / noise_gain
            mixture.noise_snr_gain = 1
        else:
            mixture.target_snr_gain = 1
            mixture.noise_snr_gain = noise_gain

    # Check for clipping in mixture
    gain_adjusted_target_audio = target_audio * mixture.target_snr_gain
    gain_adjusted_noise_audio = np.single(noise_audio) * mixture.noise_snr_gain
    mixture_audio = gain_adjusted_target_audio + gain_adjusted_noise_audio
    if any(abs(mixture_audio) >= 32768):
        # Clipping occurred; lower gains to bring audio within int16 bounds
        gain_adjustment = dbfs_to_sample(-0.25) / max(abs(mixture_audio))
        mixture.target_snr_gain *= gain_adjustment
        mixture.noise_snr_gain *= gain_adjustment

    return mixture



from typing import Any


def import_keras_model(name: str) -> Any:
    import os
    import sys
    from importlib import import_module

    from sonusai import SonusAIError

    try:
        path = os.path.dirname(name)
        if len(path) < 1:
            path = './'

        # Add model file location to system path
        sys.path.append(os.path.abspath(path))

        try:
            # expect at least a Keras-tuner build_model(hp) function
            root = os.path.splitext(os.path.basename(name))[0]
            model = import_module(root)
        except Exception as e:
            raise SonusAIError(f'Error: could not import model from {name}: {e}.')
    except Exception as e:
        raise SonusAIError(f'Error: could not find {name}: {e}.')

    return model



from typing import List


def expand_range(s: str, sort: bool = True) -> List[int]:
    """Returns a list of integers from a string input representing a range."""
    import re

    clean_s = s.replace(':', '-')
    clean_s = clean_s.replace(';', ',')
    clean_s = re.sub(' +', ',', clean_s)
    clean_s = re.sub(',+', ',', clean_s)

    r = []
    for i in clean_s.split(','):
        if '-' not in i:
            r.append(int(i))
        else:
            l, h = map(int, i.split('-'))
            r += range(l, h + 1)

    if sort:
        r = sorted(r)

    return r


def consolidate_range(r: List[int]) -> str:
    """Returns a string representing a range from an input list of integers."""

    def ranges(i: List[int]) -> str:
        import itertools

        for a, b in itertools.groupby(enumerate(i), lambda pair: pair[1] - pair[0]):
            b = list(b)
            yield b[0][1], b[-1][1]

    ls = list(ranges(r))
    for idx, val in enumerate(ls):
        ls[idx] = f'{val[0]}'
        if val[0] != val[1]:
            ls[idx] += f'-{val[1]}'

    return ', '.join(ls)



import numpy as np

from sonusai.mixture.mixdb import Mixture
from sonusai.mixture.mixdb import MixtureDatabase


def generate_segsnr(mixdb: MixtureDatabase,
                    mixture: Mixture,
                    target_audio: np.ndarray,
                    noise_audio: np.ndarray,
                    compute: bool = True,
                    frame_based: bool = False) -> np.ndarray:
    """Generate segmental SNR."""
    from pyaaware import ForwardTransform

    from sonusai.utils import int16_to_float

    if not compute:
        return np.empty(0, dtype=np.single)

    fft = ForwardTransform(N=mixdb.frame_size * 4, R=mixdb.frame_size)

    if frame_based:
        segsnr = np.empty(mixture.samples // mixdb.frame_size, dtype=np.single)
    else:
        segsnr = np.empty(mixture.samples, dtype=np.single)

    frame = 0
    for offset in range(0, mixture.samples, mixdb.frame_size):
        indices = slice(offset, offset + mixdb.frame_size)

        target_energy = fft.energy_t(int16_to_float(target_audio[indices]))
        noise_energy = fft.energy_t(int16_to_float(noise_audio[indices]))

        if noise_energy == 0:
            snr = np.single(np.inf)
        else:
            snr = np.single(target_energy / noise_energy)

        if frame_based:
            segsnr[frame] = snr
            frame += 1
        else:
            segsnr[indices] = snr

    return segsnr



import numpy as np


def calculate_sample_weights(cweights: np.ndarray, truth: np.ndarray) -> np.ndarray:
    """Calculate sample weights from class weights and a given truth with 2D or 3D shape.

    Supports one-hot encoded multi-class or binary truth/labels
    Note returns sum of weighted truth over classes, thus should also work for multi-label ? TBD

    Inputs:
      cweights num_classes x 1        weights for each class
      truth    frames x timesteps x num_classes or frames x num_classes

    Returns:
      sweights frames x timesteps x 1 or frames x 1
    """
    ts = truth.shape
    cs = cweights.shape

    if ts[-1] == 1 and cs[0] == 2:
        # Binary truth needs 2nd "none" truth dimension
        truth = np.concatenate((truth, 1 - truth), axis=1)

    # broadcast num_classes x 1 over frames x num_classes or frames x timesteps x num_classes
    sweights = np.sum(cweights * truth, axis=-1)
    return sweights



"""sonusai keras_train

usage: keras_train [-hvfc] (-m MODEL) [-e EPOCHS] [-b BATCH] [-t TSTEPS] [-l VAL] DATA

options:
   -h, --help
   -v, --verbose                Be verbose.
   -f, --flatten                Flatten input feature data.
   -c, --add1ch                 Add channel dimension to feature (i.e., cnn input).
   -m MODEL, --model MODEL      Model Python file with build and/or hypermodel functions.
   -e EPOCHS --epochs EPOCHS    Number of epochs to use in training. [default: 8].
   -b BATCH, --batch BATCH      Batch size. [default: 32].
   -t TSTEPS, --tsteps TSTEPS   Timesteps.  [default: 0].
   -l VAL, --val VAL            Validation split (0 - 1) or validation HDF5 file. [default: 0.2].

Use Keras to train a model defined by a Python definition file and DATA HDF5 feature+truth file
generated from the SonusAI genft function.

Validation is loaded from VAL if HDF5 file or created from DATA using a split of VAL.

Results are written into subdirectory <MODEL>-<TIMESTAMP>

"""
from sonusai import logger


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    import os
    from datetime import datetime

    import h5py
    import keras2onnx
    import keras_tuner as kt
    import numpy as np
    import tensorflow as tf
    from tensorflow.keras import backend as kb
    from tensorflow.keras.callbacks import EarlyStopping

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.data_generator import DataGenerator
    from sonusai.genft import genft
    from sonusai.metrics import class_summary
    from sonusai.metrics import snr_summary
    from sonusai.mixture import get_ft_from_file
    from sonusai.mixture import get_total_class_count
    from sonusai.mixture import load_mixdb
    from sonusai.mixture import new_mixdb_from_mixid
    from sonusai.utils import calculate_input_shape
    from sonusai.utils import create_onnx_from_keras
    from sonusai.utils import import_keras_model
    from sonusai.utils import reshape_inputs
    from sonusai.utils import stratified_shuffle_split_mixid

    verbose = args['--verbose']
    flatten = args['--flatten']
    add1ch = args['--add1ch']
    model_name = args['--model']
    epochs = int(args['--epochs'])
    batch_size = int(args['--batch'])
    timesteps = int(args['--tsteps'])
    val_name = args['--val']
    data_name = args['DATA']

    model_tail = os.path.basename(model_name)
    model_root = os.path.splitext(model_tail)[0]

    # Results subdirectory
    ts = datetime.now()

    # First try just date
    ts_str = model_root + '-' + ts.strftime('%Y%m%d')
    try:
        os.mkdir(ts_str)
    except OSError as _:
        # add hour-min-sec if necessary
        ts_str = model_root + '-' + ts.strftime('%Y%m%d-%H%M%S')
        os.mkdir(ts_str)

    # Setup logging file
    name_ts = ts_str + '/' + model_root  # base filename path with subdir timestamp
    create_file_handler(name_ts + '.log')
    update_console_handler(verbose)
    initial_log_messages(model_root)

    # Check dims and model build before we read large dataset and make subdir and start logging
    logger.info(f'TF ver: {tf.__version__}')
    logger.info(f'Keras ver: {tf.keras.__version__}')
    logger.info(f'Keras2onnx ver: {keras2onnx.__version__}')

    mixdb = load_mixdb(data_name)
    num_mixtures = len(mixdb.mixtures)
    logger.info(f'Found {num_mixtures} mixtures with {mixdb.num_classes} classes from {data_name}')

    # Import model definition file
    model = import_keras_model(model_name)
    logger.info(f'Successfully imported {model_tail}, testing default model build')

    # Calculate input shape
    logger.info('Building default model')

    try:
        hypermodel = model.MyHyperModel(num_classes=mixdb.num_classes,
                                        feature=mixdb.feature,
                                        timesteps = timesteps,
                                        flatten = flatten,
                                        add1ch = add1ch,
                                        batch_size=batch_size)
        model_default = hypermodel.build_model(kt.HyperParameters())
    except Exception as e:
        logger.exception(f'Error: build_model() in {model_tail} failed: {e}.')
        raise SystemExit(1)

    logger.info(f'Successfully built using {model_tail}, summary:')
    kb.clear_session()
    logger.info('')
    model_default.summary(print_fn=logger.info)
    logger.info(f'User shape parameters: batch_size {batch_size}, timesteps {timesteps}, '
                f'flatten={flatten}, add1ch={add1ch}')
    logger.info(f'Model build above with default hyper-parameters, in_shape: {hypermodel.input_shape}, '
                f'num_classes {mixdb.num_classes}')
    logger.info(f'Compiled with optimizer: {model_default.optimizer.get_config()}')
    logger.info('')

    # Prepare validation data
    v_mixdb = None
    v_feature = None
    v_truth = None
    vsplit = None
    try:
        vsplit = float(val_name)
        if vsplit < 0 or vsplit > 1:
            logger.exception('--val must be between 0 and 1')
            raise SystemExit(1)
    except ValueError:
        try:
            val_head, val_tail = os.path.split(val_name)
            val_ext = os.path.splitext(val_tail)[1]
            v_mixdb = load_mixdb(val_name)
            if val_ext == '.h5':
                v_feature, v_truth = get_ft_from_file(val_name)
            else:
                # JSON - generate feature-truth data
                v_feature, v_truth, _, _ = genft(mixdb=v_mixdb, mixid=':', show_progress=verbose)

        except Exception as e:
            logger.warning(f'Error {e}. Using vsplit=0.2 instead of --val={val_name}')
            vsplit = 0.2

    if v_mixdb is not None:
        # If validation data provided, then use it instead of split
        logger.info(f'Read {len(v_mixdb.mixtures)} validation mixtures from {val_name}')
        use_strat = False
        if use_strat:
            logger.info(f'Stratify/shuffle all {num_mixtures} training mixtures')
            # Training data does not need to be split; use all mixtures from mixdb
            t_mixid, _, _, _ = stratified_shuffle_split_mixid(mixdb, vsplit=0)
        else:
            t_mixid = range(num_mixtures)
    else:
        # vsplit from --val arg or default = 0.2
        logger.info(f'Calculating validation data using split of {vsplit}')
        use_strat = False
        if use_strat:
            t_mixid, v_mixid, t_num_mixid, v_num_mixid = stratified_shuffle_split_mixid(mixdb, vsplit=vsplit)
            logger.info(f'Split {len(v_mixid)} validation mixtures')
            v_feature, v_truth = get_ft_from_file(filename=data_name, mixid=v_mixid)
            v_mixdb = new_mixdb_from_mixid(mixdb, v_mixid)
        else:
            mxsplit = int(np.ceil((1-vsplit)*num_mixtures))
            t_mixid = range(mxsplit)
            v_mixid = range(mxsplit,num_mixtures)
            logger.info(f'Split {len(v_mixid)} validation mixtures without stratification.')
            v_feature, v_truth = get_ft_from_file(filename=data_name, mixid=v_mixid)
            v_mixdb = new_mixdb_from_mixid(mixdb, v_mixid)

    v_mixid = list(range(len(v_mixdb.mixtures)))
    logger.info(f'Original validation data shape: {v_feature.shape}')
    v_feature, v_truth, _, _, _, _ = reshape_inputs(feature=v_feature,
                                                    truth=v_truth,
                                                    batch_size=batch_size,
                                                    timesteps=timesteps,
                                                    flatten=flatten,
                                                    add1ch=add1ch)
    logger.info(f'Final validation data shape: {v_feature.shape}')

    # Prepare class weighting
    class_featcnt = np.ceil(np.array(get_total_class_count(mixdb, t_mixid)) / mixdb.feature_step_samples)
    if mixdb.truth_mutex:
        oweight = 16.0
        logger.info(f'Detected single-label mode (truth_mutex), setting other weight to {oweight}')
        class_featcnt[-1] = class_featcnt[-1] / oweight

    total_features = sum(class_featcnt)
    #cweights = np.zeros(mixdb.num_classes)
    #for n in range(mixdb.num_classes):
        # Avoid non-existent problem in sklearn by setting to 0
        #cweights[n] = total_features / (mixdb.num_classes * class_featcnt[n]) if class_featcnt[n] else 0

    #logger.info(f'Final class weights: {cweights}')

    t_feature = None
    t_truth = None
    t_datagen = None

    memory_enough = True
    if memory_enough:
        # Read train+truth into memory and reshape
        logger.info(f'Reading {len(t_mixid)} training mixtures into memory')
        t_feature, t_truth = get_ft_from_file(filename=data_name, mixid=t_mixid)
        logger.info(f'Reshaping training data')
        t_feature, t_truth, _, _, _, _ = reshape_inputs(feature=t_feature,
                                                        truth=t_truth,
                                                        batch_size=batch_size,
                                                        timesteps=timesteps,
                                                        flatten=flatten,
                                                        add1ch=add1ch)
        #class_truth_cnt = sum(t_truth)
        #fframes = t_truth.shape[0]
        #ctweights = np.zeros(mixdb.num_classes)
        #for n in range(mixdb.num_classes):
        #    ctweights[n] = fframes / (class_truth_cnt[n]) if class_truth_cnt[n] else 1
    else:
        # Use SonusAI DataGenerator to create training+truth on the fly
        logger.info(f'Split {len(t_mixid)} training mixtures')
        t_datagen = DataGenerator(filename=data_name,
                                  mixid=t_mixid,
                                  batch_size=batch_size,
                                  timesteps=timesteps,
                                  flatten=flatten,
                                  add1ch=add1ch)

    es_p = 8
    es = EarlyStopping(monitor='val_loss',
                       mode='min',
                       verbose=1,
                       patience=es_p)
    ckpt_callback = tf.keras.callbacks.ModelCheckpoint(filepath=name_ts + '-ckpt-weights.h5',
                                                       save_weights_only=True,
                                                       monitor='val_loss',
                                                       mode='min',
                                                       save_best_only=True)

    logger.info('')
    logger.info(f'Training with no class weighting and early stopping patience = {es_p}')
    logger.info('')

    if memory_enough:
        history = model_default.fit(t_feature, t_truth,
                                    batch_size=batch_size,
                                    epochs=epochs,
                                    validation_data=(v_feature, v_truth),
                                    shuffle=False,
                                    callbacks=[es, ckpt_callback])
    else:
        history = model_default.fit(t_datagen,
                                    batch_size=batch_size,
                                    epochs=epochs,
                                    validation_data=(v_feature, v_truth),
                                    shuffle=False,
                                    callbacks=[es, ckpt_callback])

    # Save history into numpy file
    history_name = name_ts + '-history'
    np.save(history_name, history.history)
    # Note: Reload with history=np.load(history_name, allow_pickle='TRUE').item()
    logger.info(f'Saved training history to numpy file {history_name}.npy')

    # Find checkpoint file and load weights for prediction and model save
    ckptfname = None
    for path, dirs, files in os.walk(ts_str):
        for f in files:
            if "ckpt" in f:
                ckptfname = f

    if ckptfname is not None:
        logger.info('Using best checkpoint for prediction and model exports')
        weight_fname = ts_str + '/' + ckptfname  # save for later model export(s)
        model_default.load_weights(weight_fname)
        model_default.save(name_ts + '.h5')
    else:
        logger.info('Using last epoch for')
        model_default.save(name_ts + '.h5')
        weight_fname = name_ts + '.h5'   # save for later model export(s)

    logger.info(f'Saved trained model to {name_ts}.h5')
    # Save to ONNX format with specified batch and timestep sizes
    try:
        create_onnx_from_keras(keras_model=model_default,
                               is_flattened=flatten,
                               has_timestep=(timesteps != 0),
                               has_channel=add1ch,
                               is_mutex=mixdb.truth_mutex,
                               feature=mixdb.feature,
                               filename=name_ts + '.onnx')
    except Exception as e:
        logger.warning(f'Failed to create ONNX model, no file saved: {e}.')
    logger.info('Wrote model to onnx file {}'.format(name_ts + '.onnx'))

    # Compute prediction metrics on validation data using saved best checkpoint
    # t_mixid, v_mixid, or entire dataset mixdb
    v_predict = model_default.predict(x=v_feature, batch_size=batch_size, verbose=1)
    with h5py.File(name_ts + '-valpredict.h5', 'w') as f:
        f.create_dataset('predict', data=v_predict)
        f.create_dataset('truth_f', data=v_truth)
        f.attrs['mixdb'] = v_mixdb.to_json()

    logger.info('Wrote prediction and truth to file {}'.format(name_ts + '-valpredict.h5'))
    # Create and save onnx model with timesteps, batch = 1
    #logger.info('')
    if timesteps > 0:
        # only set to 1 if nonzero (exists)
        timesteps = 1
    hypermodelp = model.MyHyperModel(num_classes=mixdb.num_classes,
                                    feature=mixdb.feature,
                                    timesteps=timesteps,
                                    flatten=flatten,
                                    add1ch=add1ch,
                                    batch_size=1)

    modelp = hypermodelp.build_model(kt.HyperParameters())
    # load weights from previously saved HDF5
    modelp.load_weights(weight_fname)
    # save a prediction version of model to name_ts-pred-onnx
    create_onnx_from_keras(keras_model=modelp,
                           is_flattened=flatten,
                           has_timestep=(timesteps != 0),
                           has_channel=add1ch,
                           is_mutex=mixdb.truth_mutex,
                           feature=mixdb.feature,
                           filename=name_ts + '-b1.onnx')
    logger.info('Wrote inference model (batch_size=1) to onnx file {}'.format(name_ts + '-b1.onnx'))

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



from typing import List

import numpy as np

from sonusai.mixture.mixdb import Mixture
from sonusai.mixture.mixdb import MixtureDatabase


def _strictly_decreasing(list_to_check: list) -> bool:
    return all(x > y for x, y in zip(list_to_check, list_to_check[1:]))


def generate_truth(mixdb: MixtureDatabase,
                   mixture: Mixture,
                   target_audios: List[np.ndarray],
                   noise_audio: np.ndarray,
                   compute: bool = True) -> np.ndarray:
    from sonusai import SonusAIError

    if not compute:
        return np.empty(0, dtype=np.single)

    if len(target_audios) != len(mixture.target_file_index):
        raise SonusAIError('Number of target audio entries does not match number of target file indices')

    if not all(len(target_audio) == len(noise_audio) for target_audio in target_audios):
        raise SonusAIError('Lengths of target audio do not match length of noise audio')

    truth = np.zeros((mixture.samples, mixdb.num_classes), dtype=np.single)
    for idx in range(len(target_audios)):
        for truth_setting in mixdb.targets[mixture.target_file_index[idx]].truth_settings:
            config = {
                'index':       truth_setting.index,
                'function':    truth_setting.function,
                'config':      truth_setting.config,
                'frame_size':  mixdb.frame_size,
                'num_classes': mixdb.num_classes,
                'mutex':       mixdb.truth_mutex,
                'target_gain': mixture.target_gain[idx] * mixture.target_snr_gain
            }
            truth += truth_function(target_audio=target_audios[idx], noise_audio=noise_audio, config=config)

    return truth


def truth_function(target_audio: np.ndarray,
                   noise_audio: np.ndarray,
                   config: dict) -> np.ndarray:
    import h5py
    from pyaaware import ForwardTransform
    from pyaaware import SED

    from sonusai import SonusAIError
    from sonusai.mixture.mapped_snr_f import calculate_mapped_snr_f
    from sonusai.utils.int16_to_float import int16_to_float

    truth = np.zeros((len(target_audio), config['num_classes']), dtype=np.single)
    zero_based_indices = [x - 1 for x in config['index']]

    if config['function'] == 'sed':
        if len(target_audio) % config['frame_size'] != 0:
            raise SonusAIError(f'Number of samples in audio is not a multiple of {config["frame_size"]}')

        if 'config' not in config:
            raise SonusAIError('Truth function SED missing config')

        parameters = ['thresholds']
        for parameter in parameters:
            if parameter not in config['config']:
                raise SonusAIError(f'Truth function SED config missing required parameter: {parameter}')

        thresholds = config['config']['thresholds']
        if not _strictly_decreasing(thresholds):
            raise SonusAIError(f'Truth function SED thresholds are not strictly decreasing: {thresholds}')

        if config['target_gain'] == 0:
            return truth

        target_fft = ForwardTransform(N=config['frame_size'] * 4, R=config['frame_size'])
        # SED wants 1-based indices
        sed = SED(thresholds=thresholds,
                  index=config['index'],
                  frame_size=config['frame_size'],
                  num_classes=config['num_classes'],
                  mutex=config['mutex'])

        target_audio = np.int16(np.round(np.single(target_audio) / config['target_gain']))
        for offset in range(0, len(target_audio), config['frame_size']):
            indices = slice(offset, offset + config['frame_size'])
            new_truth = sed.execute(target_fft.energy_t(int16_to_float(target_audio[indices])))
            truth[indices] = np.reshape(new_truth, (1, len(new_truth)))

        return truth

    elif config['function'] == 'file':
        if 'config' not in config:
            raise SonusAIError('Truth function file missing config')

        parameters = ['file']
        for parameter in parameters:
            if parameter not in config['config']:
                raise SonusAIError(f'Truth function file config missing required parameter: {parameter}')

        with h5py.File(name=config['config']['file'], mode='r') as f:
            truth_in = np.array(f['/truth_t'])

        if truth_in.ndim != 2:
            raise SonusAIError('Truth file data is not 2 dimensions')

        if truth_in.shape[0] != len(target_audio):
            raise SonusAIError('Truth file does not contain the right amount of samples')

        if config['target_gain'] == 0:
            return truth

        if len(zero_based_indices) > 1:
            if len(zero_based_indices) != truth_in.shape[1]:
                raise SonusAIError('Truth file does not contain the right amount of classes')

            truth[:, zero_based_indices] = truth_in
        else:
            index = zero_based_indices[0]
            if index + truth_in.shape[1] > config['num_classes']:
                raise SonusAIError('Truth file contains too many classes')

            truth[:, index:index + truth_in.shape[1]] = truth_in

        return truth

    elif config['function'] == 'energy_t':
        if config['target_gain'] == 0:
            return truth

        target_fft = ForwardTransform(N=config['frame_size'] * 4, R=config['frame_size'])
        for offset in range(0, len(target_audio), config['frame_size']):
            target_energy = target_fft.energy_t(int16_to_float(target_audio[offset:offset + config['frame_size']]))
            truth[offset:offset + config['frame_size'], zero_based_indices] = np.single(target_energy)

        return truth

    elif config['function'] in ['energy_f', 'snr_f', 'mapped_snr_f']:
        if config['target_gain'] == 0:
            return truth

        target_fft = ForwardTransform(N=config['frame_size'] * 4, R=config['frame_size'])
        noise_fft = ForwardTransform(N=config['frame_size'] * 4, R=config['frame_size'])

        snr_db_mean = None
        snr_db_std = None
        if config['function'] == 'mapped_snr_f':
            if 'config' not in config:
                raise SonusAIError('Truth function mapped SNR missing config')

            parameters = ['snr_db_mean', 'snr_db_std']
            for parameter in parameters:
                if parameter not in config['config']:
                    raise SonusAIError(f'Truth function mapped SNR config missing required parameter: {parameter}')

            snr_db_mean = config['config']['snr_db_mean']
            if len(snr_db_mean) != target_fft.bins:
                raise SonusAIError(f'Truth function mapped SNR snr_db_mean does not have {target_fft.bins} elements')

            snr_db_std = config['config']['snr_db_std']
            if len(snr_db_std) != target_fft.bins:
                raise SonusAIError(f'Truth function mapped SNR snr_db_std does not have {target_fft.bins} elements')

        for index in zero_based_indices:
            if index + target_fft.bins > config['num_classes']:
                raise SonusAIError('Truth index exceeds the number of classes')

        for offset in range(0, len(target_audio), config['frame_size']):
            target_energy = np.single(
                target_fft.energy_f(int16_to_float(target_audio[offset:offset + config['frame_size']])))
            if config['function'] in ['snr_f', 'mapped_snr_f']:
                noise_energy = np.single(
                    noise_fft.energy_f(int16_to_float(noise_audio[offset:offset + config['frame_size']])))
            else:
                noise_energy = np.single(1)

            indices = slice(offset, offset + config['frame_size'])
            for index in zero_based_indices:
                old_err = np.seterr(divide='ignore', invalid='ignore')
                tmp = target_energy / noise_energy
                np.seterr(**old_err)

                tmp = np.nan_to_num(tmp, nan=-np.inf, posinf=np.inf, neginf=-np.inf)

                if config['function'] == 'mapped_snr_f':
                    tmp = calculate_mapped_snr_f(tmp, snr_db_mean, snr_db_std)

                truth[indices, index:index + target_fft.bins] = tmp

        return truth

    elif config['function'] == 'phoneme':
        # Read in .txt transcript and run a Python function to generate text grid data
        # (indicating which phonemes are active)
        # Then generate truth based on this data and put in the correct classes based on config['index']
        raise SonusAIError('Truth function phoneme is not supported yet')

    raise SonusAIError(f'Unsupported truth function: {config["function"]}')


def get_truth_indices_for_mixid(mixdb: MixtureDatabase, mixid: int) -> List[int]:
    """Get a list of truth indices for a given mixid."""

    from sonusai.mixture.targets import get_truth_indices_for_target

    indices = list()
    for target_file_index in mixdb.mixtures[mixid].target_file_index:
        indices.append(*get_truth_indices_for_target(mixdb.targets[target_file_index]))

    return sorted(list(set(indices)))


def truth_reduction(x: np.ndarray, func: str) -> np.ndarray:
    from sonusai import SonusAIError

    if func == 'max':
        return np.max(x, axis=0)

    if func == 'mean':
        return np.mean(x, axis=0)

    raise SonusAIError(f'Invalid truth reduction function: {func}')
